"""
This module provides a SignalView component for the app.
"""
import datetime
import streamlit as st
from txp.common import plotter_utils
from txp.devices.drivers.mock.mock_driver import MockDriver, get_signals_per_dimension
import txp.web.core_components.app_profiles import AppProfile
import txp.web.core_components.main_view import MainView
import txp.cloud.controllers.firestore as firestore_controller
from txp.common.edge import EdgeDescriptor
from txp.cloud import cloud_utils
from typing import Dict
import logging
from txp.common.config import settings
import matplotlib.pyplot as plt
from google.cloud import bigquery
import numpy as np
import cv2
import pytz

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


class SignalView(MainView):
    """A Signal view that helps to test and see tasks data"""

    def __init__(self):

        self.fft_table = ".telemetrydataset01.fft"
        self.samples_table = ".telemetrydataset01.samples"
        super(SignalView, self).__init__(
            component_key="signal_view"
        )

    def get_associated_profile(self) -> AppProfile:
        return AppProfile.ANALYTICS

    @classmethod
    def get_display_name(cls) -> str:
        return 'Ver señales'

    def _build_submenu_state(self) -> Dict:
        # Hardcoded state for demonstration
        return {
            "edge_logical_id": None
        }

    def _build_content_state(self) -> Dict:
        return {}

    def _render_submenu(self) -> None:
        with st.form("date_form"):
            date = st.date_input("Pick a date", datetime.datetime.now(pytz.timezone("America/Mexico_City")))
            submitted_date = st.form_submit_button("refresh")

        if submitted_date:
            date += datetime.timedelta(days=1)
            date_time = datetime.datetime(date.year, date.month, date.day, tzinfo=pytz.timezone("America/Mexico_City"))
            self.pull_project_data(date_time, self.app_state.authentication_info.role_service_account_credentials)
            if "project_model" in st.session_state and st.session_state.project_model:
                st.session_state.project_model["query_data"] = {}
                st.session_state.project_model["selected_date"] = date_time

        if "project_model" in st.session_state and st.session_state.project_model:
            with st.form("task_form"):

                selected_date = st.session_state.project_model["selected_date"]
                selected_date -= datetime.timedelta(days=1)

                task_list = []
                for i, t in enumerate(st.session_state["project_model"]["job"]["tasks"]):
                    start = datetime.datetime.strptime(t["parameters"]["start_date"], '%Y-%m-%d')
                    end = datetime.datetime.strptime(t["parameters"]["end_date"], '%Y-%m-%d')
                    start = datetime.datetime(start.year, start.month, start.day,
                                              tzinfo=pytz.timezone("America/Mexico_City"))
                    end = datetime.datetime(end.year, end.month, end.day, tzinfo=pytz.timezone("America/Mexico_City"))
                    if start <= selected_date <= end:
                        task_list.append(f'{i}, from {t["parameters"]["start_date"]} to {t["parameters"]["end_date"]}')

                task = st.selectbox("Tasks", task_list)
                submitted_tasks = st.form_submit_button("refresh")
            if submitted_tasks:
                st.session_state.project_model["query_data"] = {}
                st.session_state.project_model["query_data"]["task"] = int(task.split(",")[0])

            with st.form("motor_form"):
                machine_list = []
                if "task" in st.session_state.project_model["query_data"]:
                    task = st.session_state.project_model["query_data"]["task"]
                    machine_list = [machine for machine in
                                    st.session_state["project_model"]["job"]["tasks"][task]["machines"]]
                machine = st.selectbox(
                    "Machine",
                    machine_list
                )
                submitted_machines = st.form_submit_button("refresh")
            if submitted_machines:
                st.session_state.project_model["query_data"].pop("edge", None)
                st.session_state.project_model["query_data"]["machine"] = machine

            with st.form("edge_form"):
                edge_list = []
                if "machine" in st.session_state.project_model["query_data"]:
                    machine_id = st.session_state.project_model["query_data"]["machine"]
                    machine_doc_ref = None
                    for machine in st.session_state.project_model['machines']:
                        if machine['machine_id'] == machine_id:
                            machine_doc_ref = machine
                            break

                    edge_list = list(map(
                        lambda edge_doc_ref: edge_doc_ref.get().get("logical_id"),
                        machine_doc_ref["associated_with_edges"]
                    ))
                edge = st.selectbox(
                    "edge",
                    edge_list
                )
                submitted_edges = st.form_submit_button("query")
            if submitted_edges:
                st.session_state.project_model["query_data"]["edge"] = edge

            with st.form("ok_label_form"):
                ok_label_set = False
                if "signal" in st.session_state.project_model:
                    element = st.session_state.project_model["signal"]
                    if element["label"] == "OK":
                        ok_label_set = True
                ok_label = st.checkbox("OK", ok_label_set)
                submitted = st.form_submit_button("Set label")
            if submitted:
                if "signal" in st.session_state.project_model:
                    label = "BAD"
                    if ok_label:
                        label = "OK"
                    credentials = self.app_state.authentication_info.role_service_account_credentials
                    project_id = credentials.project_id
                    client = bigquery.Client(credentials=credentials, project=project_id)
                    cloud_utils.set_label_for_signal(label, element, project_id + self.samples_table, client)

            if "edge" in st.session_state.project_model["query_data"]:
                self._submenu_state['edge_logical_id'] = st.session_state.project_model["query_data"]["edge"]
            else:
                self._submenu_state['edge_logical_id'] = None

    def _render_content(self) -> None:
        if "project_model" in st.session_state and st.session_state.project_model:
            st.markdown(f"#### {st.session_state.project_model['machines_groups'][0]['name']}")
            st.markdown("##### Última configuración")
            self._render_configuration()

            # Render BigQuery Form
            st.markdown("--------")
            st.markdown("#### Visualizar gráficas")

            with st.form("bigquery_form"):
                perceptions = []
                for edge in st.session_state.project_model["edges"]:
                    if edge["logical_id"] == self._submenu_state['edge_logical_id']:
                        perceptions = edge["is_of_kind_ref"].get().get("perceptions").keys()
                        break
                perception_name = st.selectbox(
                    "Tipo de percepción",
                    perceptions
                )

                logaritmic_scale = st.checkbox("Escala Logarítmica")
                number = st.number_input('X ticks, scala', min_value=-1)  # -1 means no scale given

                submitted = st.form_submit_button("Mostrar Grafica")
            project_id = self.app_state.authentication_info.role_service_account_credentials.project_id
            if submitted:
                self.plot_signals(
                    self._submenu_state['edge_logical_id'],
                    perception_name,
                    project_id + self.fft_table,
                    project_id + self.samples_table,
                    logaritmic_scale,
                    None if number == -1 else number
                )

    # =====================Helper Functions==================================
    def get_edge_descriptors(self):
        return [EdgeDescriptor(edge["logical_id"], edge["device_kind"],
                               edge["is_of_kind_ref"].get().get("type"),
                               edge["is_of_kind_ref"].get().get("parameters"),
                               edge["is_of_kind_ref"].get().get("perceptions"))
                for edge in st.session_state.project_model["edges"]]

    def pull_project_data(self, selected_date, service_account_role_credentials):
        db = firestore_controller.get_authenticated_client(
            self.app_state.authentication_info.role_service_account_credentials
        )

        st.session_state.project_model = firestore_controller.pull_project_data_by_date(
            db, selected_date
        )

        db.close()

    def _render_machine(self, machine: Dict):
        st.write(f"**Machine:** {machine['machine_id']}")
        st.write(f"**Manufacturer:** {machine['manufacturer']}")
        st.write(f"**Model:** {machine['model']}")
        with st.expander(f"Edges:"):
            for edge_ref in machine['associated_with_edges']:
                edge_snapshot = edge_ref.get()
                st.write(f'{edge_snapshot.get("logical_id")}')

    def _render_configuration(self):
        configuration = st.session_state.project_model['last_configuration']
        st.markdown(f"**Válida desde**: {configuration.get('since')}")

    def plot_all_fft(self, fft_bigquery_per_dimension, edge, perception_name, semilogy=False, x_scale=None):
        if edge.device_kind == MockDriver.device_name():
            signals_per_dimension = get_signals_per_dimension(edge)
            for i, signal in enumerate(signals_per_dimension):
                plotter_signal = plotter_utils.FFTPlotter(signal.signal_amplitude_axis, None, signal.sampling_frequency)
                fig = plotter_signal.plot(f"Expected signal, dimension {i}", semilogy, x_scale)
                st.pyplot(fig=fig)

        for i, dimension in enumerate(fft_bigquery_per_dimension):
            sampling_frequency = edge.perceptions[perception_name]["sampling_frequency"]
            plotter_bigquery = plotter_utils.FFTPlotter(None, dimension, sampling_frequency)
            fig = plotter_bigquery.plot(
                f"From cloud data fft. Edge: {edge.logical_id}, Perception: {perception_name}, Dimension {i}",
                semilogy, x_scale)
            st.pyplot(fig=fig)

    def plot_all_time(self, data_bigquery_per_dimension, edge, perception_name, semilogy=False):
        for i, dimension in enumerate(data_bigquery_per_dimension):
            fig = plt.figure(figsize=(23, 10))
            if semilogy:
                plt.semilogy(dimension)
            else:
                plt.plot(dimension, marker=".", markersize=50)
            fig.suptitle(
                f"From cloud data signal. Edge: {edge.logical_id}, Perception: {perception_name}, Dimension {i}",
                fontsize=20)
            st.pyplot(fig=fig)

    def plot_image(self, image_from_bigquery, edge, perception_name):
        image = [int(x) for x in image_from_bigquery[0]]
        image = np.frombuffer(bytes(image), dtype=np.uint8)
        image = cv2.imdecode(image, flags=1)
        fig = plt.figure(figsize=(23, 10))
        plt.imshow(image[:, :, ::-1])
        fig.suptitle(f"From cloud data image. Edge: {edge.logical_id}, Perception: {perception_name}",
                     fontsize=20)
        st.pyplot(fig=fig)
        pass

    def plot_signals(self, edge_logical_id, perception_name, table_fft, table_samples, semilogy=False, x_scale=None):
        edge = None
        for e in self.get_edge_descriptors():
            if e.logical_id == edge_logical_id:
                edge = e
                break
        client = bigquery.Client(
            credentials=self.app_state.authentication_info.role_service_account_credentials,
            project=self.app_state.authentication_info.role_service_account_credentials.project_id
        )
        if edge is not None:
            end_date = st.session_state.project_model["selected_date"]
            start_date = end_date
            start_date -= datetime.timedelta(days=1)
            if edge.perceptions[perception_name]["mode"] == "IMAGE":
                element = cloud_utils.get_signal_from_bigquery(perception_name, edge_logical_id, start_date, end_date,
                                                               table_samples, client)
                st.session_state.project_model["signal"] = element
                self.plot_image(element["data"], edge, perception_name)

            elif edge.perceptions[perception_name]["mode"] == "FFT":
                fft_element = cloud_utils.get_fft_from_bigquery(perception_name, edge_logical_id, start_date, end_date,
                                                                table_fft, table_samples, client)
                st.session_state.project_model["signal"] = fft_element
                self.plot_all_fft(fft_element["fft"], edge, perception_name, semilogy, x_scale)
            else:
                element = cloud_utils.get_signal_from_bigquery(perception_name, edge_logical_id, start_date, end_date,
                                                               table_samples, client)
                st.session_state.project_model["signal"] = element
                self.plot_all_time(element["data"], edge, perception_name, semilogy)
        else:
            raise Exception(f"edge_logical_id: {edge_logical_id} not found at configuration")



"""
This module defines the base exceptions to be used in the Gateway
system.
"""
# ============================ imports =========================================
class GatewayRecoverableException(Exception):
    """This exception class is thought to use across the system
    when errors that are **expected** could happen and are treated
    gracefully by the callers.
    """
    pass


class GatewayFatalException(Exception):
    """This exception base class is thought to use across the system
    when an error should transition the Gateway States machine to an
    error state."""
    pass


class MqttNetworkingUnavailableError(GatewayRecoverableException):
    """This exception should be used when a unexpected error
    is found with the networking of the device."""
    pass


class MqttPrivateKeyError(GatewayRecoverableException):
    """This exception should be used when the received private
    key provided could not be parsed."""
    pass


class MqttConnectTimeoutError(GatewayRecoverableException):
    """This exception should be used when the MQTT paho client
    can not establish the connection with the broker. It's most
    used for internet outages."""
    pass


class MqttMessageNotPublished(GatewayRecoverableException):
    """This exception should be used when the MQTT paho client
    fails to publish a message to the broker.

    Specifically, the rc code is not zero:
        https://www.eclipse.org/paho/index.php?page=clients/python/docs/index.php#publishing
    """
    pass



"""
This module contains the concrete driver implementation for the Thermal Robot device.
"""

# =================== imports ==============================================
import queue
import threading
import time
import logging
from typing import Dict, Tuple, List, Callable

from txp.devices.drivers.arm_robot.arm_controller import ArmController
from txp.devices.drivers.arm_robot.arm_handler import ArmHandler
from txp.devices.drivers.usb.camera.camera_handler_base import CameraHandlerBase, CameraType
from txp.devices.drivers.driver import Driver, EdgeParameterNotFoundError, DriverState
from txp.common.edge import EdgeDescriptor, VirtualSensorObject, VirtualEdgeInformation, TemperatureSignal
from txp.devices.package_queue import PackageQueueProxy
import mcp9600
from txp.common.config import settings

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


# ==============================================================================
# Thermal Robot driver implementation
# ==============================================================================
class ThermalRobotDriver(Driver):
    """Implementation of the Arm Robot driver used with the NVIDIA Jetson computer.

    The ArmRobotDriver is the main unit composed of different physical sensors
    and a set of positions which are interpreted as Virtual Edges that captures
    the perceptions of those edges.
    """

    def __init__(
            self,
            edge_descriptor: EdgeDescriptor,
            gateway_packages_queue: PackageQueueProxy,
            on_driver_state_changed: Callable,
            robotic_arm_controller: ArmController,
            host_clock_notification_queue: queue.Queue
    ):
        """
        Args:
            edge_descriptor: The EdgeDescriptor for this driver that contains all the
                information for the configured sensors and the virtual positions.
            gateway_packages_queue: The Gateway Package Queue to put produced packages.
            on_driver_state_changed: Function used to notify driver state changes in the
                upper layers.
            robotic_arm_handler: The Robotic arm controller shared across all the ArmRobotDriver
                instances
        """
        super(ThermalRobotDriver, self).__init__(edge_descriptor, gateway_packages_queue,
                                             on_driver_state_changed, host_clock_notification_queue)

        # Validations of the received descriptor
        self._validate_edge_descriptor(edge_descriptor)

        if not self._validate_position_and_virtual_id(edge_descriptor):
            raise ValueError(f"The received virtual positions in driver {self.__class__.__name__} "
                             f"do not match the received virtual IDs")

        if not self._validate_cameras(edge_descriptor.get_virtual_sensors()):
            raise ValueError(f"Unknown camera devices in the configuration for "
                             f"Virtual Edge driver {self.__class__.__name__}")

        # Validates Thermocouple
        tt = mcp9600.MCP9600()
        thermocouple_type = tt.get_thermocouple_type()
        if thermocouple_type in {'K', 'J', 'T', 'N', 'S', 'E', 'B' or 'R'}:
            log.info(f'Thermocouple validated with type {thermocouple_type} for edge {self.logical_id}')
            self._tt = tt
            self._tt.set_thermocouple_type('T')
            if self._tt.get_thermocouple_type() != 'T':
                log.error(f"Thermocouple type is not T expected but: {self._tt.get_thermocouple_type()}")
            else:
                log.info(f"Thermocouple type configured to {self._tt.get_thermocouple_type()}")
        else:
            log.error(f'Thermocouple type unexpected: {thermocouple_type} for driver {self.logical_id}')
            self._tt = None

        self._robotic_arm_controller: ArmController = robotic_arm_controller
        self._virtual_information: VirtualEdgeInformation = edge_descriptor.get_virtual_edge_information()

    def observation_pulse(self) -> int:
        """The observation pulse for Virtual Drivers.

        Note: this time interval doesn't really represent the
        driver's clock value. The reason is that, currently by design
        simplification, the WindowPolicyManager will expect 1 sample
        per virtual edge on each sampling window.

        TODO: The driver could change it's observation pulse value given
            the current sampling window, in order to reflect that design
            simplification.

        """
        return 20

    @property
    def parent_logical_id(self) -> str:
        return EdgeDescriptor.split_virtual_id(self.logical_id)[0]

    @property
    def get_virtual_id_hash(self) -> str:
        return self._virtual_information.virtual_id_hash

    def _validate_edge_descriptor(self, edge_descriptor: EdgeDescriptor):
        if not edge_descriptor.get_virtual_sensors():
            log.error(f"{self.__class__.__name__} did not receive virtual "
                      f"sensors information in EdgeDecriptor")
            raise EdgeParameterNotFoundError(f"{self.__class__.__name__} did not receive virtual "
                                             f"sensors information in EdgeDecriptor")

        if not edge_descriptor.get_virtual_edge_information():
            log.error(f"f{self.__class__.__name__} did not received virtual edge information"
                      f"in Edge Descriptor parameters")
            raise EdgeParameterNotFoundError(f"{self.__class__.__name__} did not receive virtual "
                                             f"sensors information in EdgeDecriptor")

    @staticmethod
    def _validate_position_and_virtual_id(virtual_devices: EdgeDescriptor) -> bool:
        computed_virtual_id = EdgeDescriptor.get_hashed_virtual_id(
            virtual_devices.get_virtual_edge_information().position
        )
        are_equals = computed_virtual_id == virtual_devices.get_virtual_edge_information().virtual_id_hash
        if not are_equals:
            log.error(f"Computed virtual ID does not equals received virtual ID for "
                      f"positions: {virtual_devices.get_virtual_edge_information().position}")
        return are_equals

    def _validate_cameras(self, camera_sensors: List[VirtualSensorObject]) -> bool:
        valid = True
        for camera in camera_sensors:
            if not isinstance(camera.camera_type, CameraType):
                log.error(f"{self.__class__.__name__} received unknown camera type: {camera.camera_type}")
                valid = False

            # TODO: Here we should check if the camera is valid. Handler should offer method.

        return valid

    def connect_device(self):
        log.info("conecting driver {}".format(self.logical_id))
        if self._robotic_arm_controller is not None:
            if not self._robotic_arm_controller.is_arm_fully_connected():
                self.on_driver_state_changed(
                    self.logical_id, DriverState.DISCONNECTED
                )
            else:
                self.on_driver_state_changed(
                    self.logical_id, DriverState.CONNECTED
                )
        else:
            log.info("not valid instanse of controller driver {}".format(
                self.logical_id
            ))


    def disconnect_device(self):
        self.on_driver_state_changed(
            self.logical_id, DriverState.DISCONNECTED
        )

    @classmethod
    def device_name(cls) -> str:
        return "ThermoArm"

    @classmethod
    def is_virtual(cls) -> bool:
        return True

    def _start_sampling_collect(
            self
    ) -> List["Signal"]:
        executed = False
        signals = []
        while not executed:
            if self._stop_current_pulse_signal.is_set():
                break

            if not executed:
                position = self._virtual_information.position

                signals = self._robotic_arm_controller.execute_arm_action(position)

                if signals:
                    # At this point we do have the captured images,
                    # so we capture the temperature and finish
                    temperature = self._tt.get_hot_junction_temperature()
                    if temperature is not None:
                        signal = TemperatureSignal.build_signal_from_value(
                            temperature,
                            0
                        )
                        log.info(f"Temperature capture in {self.logical_id}: {temperature}")  # TODO: Debug
                        signals.append(signal)
                    else:
                        log.error('Error reading thermocouple connected mcp9600 ')

                    executed = True

        log.info("{} finished sampling with the amr".format(self.logical_id))
        return signals

    def _stop_sampling_collect(self):
        log.info("Virtual driver {} is stopping sampling and collect".format(self.logical_id))



import sys
import os

sys.path.insert(0, os.path.join(sys.path[0], '..'))

import threading
import socket
import select
import time

from SerialConnector.Hdlc import Hdlc

from   SmartMeshSDK.ApiException  import ConnectionError, \
                                         APIError,        \
                                         CommandError
from   SmartMeshSDK.ApiConnector  import ApiConnector
from   SmartMeshSDK.ApiDefinition import ApiDefinition

import logging
class NullHandler(logging.Handler):
    def emit(self, record):
        pass
log = logging.getLogger('SerialConnector')
log.setLevel(logging.ERROR)
log.addHandler(NullHandler())

class SerialConnector(ApiConnector):
    '''
    \ingroup ApiConnector
    
    \brief The generic serial connector.
    This class is meant to be inherited by connector using a serial link.
    '''
    
    MAX_NUM_RETRY = 5
    RX_TIMEOUT    = 0.500 # in seconds
    
    def __init__(self, api_def, maxQSize=100) :
        
        # log
        log.info("creating object")
        
        # store params
        self.api_def         = api_def
        
        # initialize parent class
        ApiConnector.__init__(self, maxQSize)
        
        # local variables
        self.waitForResp     = False                  ##< flag to indicate whether we are waiting for an ACK
        self.waitForRespEvent  = threading.Event()    ##< semaphore used when waiting for a command to be acknowledged
        self.paramLock       = threading.Lock()       ##< lock to ensure atomic access to parameters
        self.RxPacketId      = None                   ##< running id for packets received from device
        self.TxPacketId      = 0                      ##< running id for packets sent to device
        self.hdlcLock        = threading.Lock()       ##< lock for the HDLC module
        self.hdlc            = None                   ##< HDLC module (includes CRC)
        self.syncNeeded      = True                   ##< True iff we need to sync to the device
        self.requestSendLock = threading.Lock()       ##< lock to prevent concurrent requests to be sent
        self.tsDataSent      = 0                      ##< timestamp when sent data
        self.tsDataReceived  = 0                      ##< timestamp when received data
        
    #======================== public ==========================================
    
    def connect(self, connectParams):
        if 'port' not in connectParams:
            output = "'port' entry required in connection parameters"
            log.error(output)
            raise ValueError(output)
        
        with self.hdlcLock:
            # create and start HDLC module (includes CRC)
            self.hdlc            = Hdlc(self._hdlcRxCb,
                                        self._hdlcConnectCb)
            # connect HDLC module to serial Port
            if 'baudrate' in connectParams:
                self.hdlc.connect(connectParams['port'],baudrate=connectParams['baudrate'])
            else:
                self.hdlc.connect(connectParams['port'])
            # connect the parent class
            ApiConnector.connect(self)
    
    def disconnect(self, reason=""):
        with self.hdlcLock:
            # disconnect the parent class
            ApiConnector.disconnect(self, reason)
            # disconnect hdlc module
            if self.hdlc:
                self.hdlc.disconnect()
            # delete the hdlc module
            self.hdlc = None
    
    def send(self,commandArray,fields):
        if not self.isConnected:
            output = "not connected"
            log.error(output)
            raise ConnectionError(output)
        
        # serialize the fields
        cmdId, serializedFields = self.api_def.serialize(commandArray,fields)
        
        return self._sendInternal(cmdId,False,serializedFields)

    #======================== virtual methods =================================

    def isValidPacketId(self, cmdId, isResponse, packetId):
        '''
        \brief Return whether a received packet is in sequence
        '''
        raise NotImplementedError() # to be implemented by child class

    def _buildTxHeader(self, cmdId, isResponse, serializedFields):
        '''
        \brief Build the header for a TX packet.
        '''
        raise NotImplementedError() # to be implemented by child class

    def _ackIfNeeded(self,cmdId,isResponse):
        '''
        \brief Send an ACK if needed.
        '''
        raise NotImplementedError() # to be implemented by child class

    def isHelloResponse(self, cmdId):
        return False
    
    #======================== private =========================================
    
    def _parseRxHeader(self,frameRx):
        raise NotImplementedError("virtual method")
    
    def _sendInternal(self,cmdId,isResponse,serializedFields):
        
        try:
            if not isResponse:
                self.requestSendLock.acquire()
        
            # build packet to send
            packet  = []
            packet += self._buildTxHeader(cmdId,isResponse,serializedFields)
            packet += serializedFields

            retry = 0
            
            # Prepare for reliable send
            if not isResponse:
                self.waitForResp = True
                self.waitForRespEvent.clear()
                
            while True :
                if log.isEnabledFor(logging.DEBUG):
                    log.debug("_sendInternal cmdId={0} retry={1} isResponse={2} serializedFields={3}".format(
                            cmdId,
                            retry,
                            isResponse,
                            serializedFields
                        )
                    )
                
                if not isResponse:
                    if log.isEnabledFor(logging.DEBUG):
                        log.debug("---------- pcToMote DATA ({0}) ---------->".format(self.TxPacketId))
                        self.tsDataSent = time.time()
                
                # send packet through HDLC module
                with self.hdlcLock:
                    if not self.hdlc:
                        output = "no HDLC module, did I just disconnect?"
                        log.error(output)
                        raise ConnectionError(output)
                    self.hdlc.send(packet)
                
                if isResponse:
                    return None
                
                # wait for response. semaphore released by _hdlcRxCb()
                if self.waitForRespEvent.wait(self.RX_TIMEOUT) :
                    break
                
                log.info("retry {0}".format(retry))
                
                # Timeout error
                if retry >= self.MAX_NUM_RETRY :
                    output = "retried {0} times, max allowed is {1}".format(retry,self.MAX_NUM_RETRY)
                    log.error(output)
                    raise ConnectionError(output)
                retry = retry + 1
                    
            if isinstance(self.responseBuf,Exception):
                log.error("responseBuf contains exception {0}".format(self.responseBuf))
                raise self.responseBuf
            
            if  (
                    (ApiDefinition.ApiDefinition.RC in self.responseBuf) and
                    (
                        self.responseBuf[ApiDefinition.ApiDefinition.RC]!= \
                            ApiDefinition.ApiDefinition.RC_OK
                    )
                ):
                temp_name = self.api_def.idToName(
                                    ApiDefinition.ApiDefinition.COMMAND,
                                    cmdId
                                )
                temp_rc   = self.responseBuf[ApiDefinition.ApiDefinition.RC]
                temp_desc = '({0})\n{1}'.format(
                    self.api_def.fieldValueToDesc(
                        ApiDefinition.ApiDefinition.COMMAND,
                        [temp_name],
                        ApiDefinition.ApiDefinition.RC,
                        temp_rc
                    ),
                    self.api_def.rcToDescription(
                        temp_rc,
                        [temp_name],
                    ),
                )
                if temp_rc not in [11,18]: # rc==11:RC_NOT_FOUND, rc==18:RC_END_OF_LIST
                    log.warning("received RC={0} for command {1}:\n{2}".format(temp_rc,
                                                                               temp_name,
                                                                               temp_desc))
                raise APIError(temp_name,temp_rc,temp_desc)
            
            # return packet received
            return self.responseBuf
        
        finally:
            if not isResponse:
                self.requestSendLock.release()
    
    def _resetPacketIds(self):
        self.TxPacketId=0
        self.RxPacketId=0
        self.syncNeeded=True
    
    #======================== HDLC callbacks ==================================
    
    def _hdlcConnectCb(self,state):
        '''
        \brief called by HDLC when the connected state has changed.
       
        \param state True when the HDLC module is connected to a device,
                     False otherwise
        '''
        log.info("hdlc notification: connection state="+str(state))

    def _hdlcRxCb(self,frameRx):
        '''
        \brief called by HDLC when it's done receiving a complete frame.
        
        \note The frame received does not contain any of the HDLC-specific
              flags and espace characters
       
        \param frameRx The received frame, represented as a byteArray
        '''
        
        if len(frameRx)<3:
            output = "received packet too short"
            log.error(output)
            raise ConnectionError(output)
        
        cmdId,length,isResponse,packetId,payload = self._parseRxHeader(frameRx)
        
        # log
        if log.isEnabledFor(logging.DEBUG):
            log.debug("cmdId={0} length={1} isResponse={2} packetId={3} payload={4}".format(
                    cmdId,
                    length,
                    isResponse,
                    packetId,
                    payload
                )
            )
        
            if isResponse:
                log.debug("<--------- pcToMote ACK ({0}) after {1:.3f} --------".format(
                                packetId,
                                time.time()-self.tsDataSent)
                         )
            else:
                log.debug("<--------- moteToPc DATA ({0}) ----------".format(packetId))
                self.tsDataReceived = time.time()
        
        # check packetId
        (wasValidPacketId, isRepeatId, updateRxPacketId) = self.isValidPacketId(cmdId,isResponse,packetId)

        # raise error if packet id is wrong
        if not wasValidPacketId:
            output = "wrong packetId"
            log.error(output)
            raise ConnectionError(output)
        
        # update RxPacketId
        self.paramLock.acquire()
        if isResponse==False and (not isRepeatId) and updateRxPacketId:
            self.RxPacketId = packetId
        self.paramLock.release()
        
        # send ACK if necessary
        try :
            ackSent = self._ackIfNeeded(cmdId,isResponse)
        except ConnectionError as ex:
            log.error("putting disconnect notification {0}".format(ex))
            self.putDisconnectNotification(str(ex))
            raise ex
        
        # log
        if log.isEnabledFor(logging.DEBUG):
            if ackSent:
                log.debug("---------- moteToPc ACK ({0}) after {1:.3f} ------->".format(
                                self.RxPacketId,
                                time.time()-self.tsDataReceived)
                          )
                log.debug("ack sent")
            else:
                log.debug("no ack needed")
        
        if isResponse or self.isHelloResponse(cmdId):
            
            # deserialize received packet
            try:
                nameArray, fields = self.api_def.deserialize(
                                        ApiDefinition.ApiDefinition.COMMAND,
                                        cmdId,
                                        payload)
            except Exception as err:
                self.responseBuf = err
            else:
                # put received packet in local response buffer
                self.responseBuf = fields
            
            # release semaphore
            if self.waitForResp:
                self.waitForResp = False
                self.waitForRespEvent.set()
            else:
                output = "unexpected response"
                log.error(output)
                raise ConnectionError(output)
            
        else:
            if not isRepeatId :
                
                # deserialize received packet
                nameArray, fields = self.api_def.deserialize(
                                        ApiDefinition.ApiDefinition.NOTIFICATION,
                                        cmdId,
                                        payload)
                
                # put received packet in notification buffer
                self.putNotification((nameArray, fields))
        



"""The Gateway Package support.

This module contains the classes that:
    - Defines the gateway package sent to cloud by the Gateway device.
    - Support the PackageQueue used to receive Driver collected packages.
"""

# ============================ imports =========================================
from dataclasses import dataclass
from txp.common.edge import EdgeDescriptor
from txp.common.edge import SignalsTableByName, Signal
from txp.common.configuration import SamplingWindow
from typing import Union
from typing import List
import txp.common.protos.gateway_package_pb2 as package_proto


# ==============================================================================
# Definition of the GatewayPackage class
# ==============================================================================


@dataclass
class GatewayPackageMetadata:
    """

    The Package ID follows the following format:
        "{epoch timestamp}_{part_index}"

    The part_index is a consequence of packages split, in order to keep size under
    250Kb.

    Args:
        edge_descriptor: The Edge descriptor of the edge device associated with this Package.
        sampling_windows: The sampling configuration windows used by the driver when it was
            collecting the package.
    """

    edge_descriptor: EdgeDescriptor
    sampling_window: SamplingWindow
    package_id: str
    """package_id: an unique string ID for this package, with the format:
    {epoch timestamp}_{part_index}
    """
    previous_part_index: int
    """
    The integer index of the previous package in the sequence generated by the package split.
    """
    tenant_id: str
    """
    The tenant_id string value of the tenant to which the Gateway belong.
    """


@dataclass
class GatewayPackage:
    """This GatewayPackage encapsulates the information exchanged between
    the Gateway and the Cloud.

    The package contains a series of samples collected during a Sampling Time (Ts)
    period.

    Args:
        configuration_id: a string indicating the version of the config.
        signals: A list of sampled signals from an edge device.
        metadata: The package metadata.
    """
    configuration_id: str
    signals: List[Signal]
    metadata: GatewayPackageMetadata

    def get_proto(self) -> package_proto.GatewayPackageProto:
        metadata_proto = package_proto.GatewayPackageMetadataProto()
        metadata_proto.edge_descriptor.CopyFrom(self.metadata.edge_descriptor.get_proto())
        metadata_proto.sampling_window.CopyFrom(self.metadata.sampling_window.get_proto())
        metadata_proto.package_id = self.metadata.package_id
        metadata_proto.previous_part_index = self.metadata.previous_part_index
        metadata_proto.tenant_id = self.metadata.tenant_id

        signals_proto = list(map(
            lambda signal: signal.get_proto(),
            self.signals
        ))
        proto = package_proto.GatewayPackageProto(configuration_id=self.configuration_id)
        proto.metadata.CopyFrom(metadata_proto)
        proto.signals.extend(signals_proto)
        return proto

    @classmethod
    def build_from_proto(cls, proto: package_proto.GatewayPackageProto) -> "GatewayPackage":
        edge_descriptor = EdgeDescriptor.build_from_proto(proto.metadata.edge_descriptor)
        sampling_window = SamplingWindow.build_from_proto(proto.metadata.sampling_window)
        package_id = proto.metadata.package_id
        previous_part_index = proto.metadata.previous_part_index
        tenant_id = proto.metadata.tenant_id
        metadata = GatewayPackageMetadata(edge_descriptor, sampling_window,
                                          package_id, previous_part_index, tenant_id)
        signals: List = []
        for signal_proto in proto.signals:
            signal_name = str(signal_proto.perception_name)
            signal = SignalsTableByName(signal_name).signal.build_from_proto(signal_proto)
            signals.append(signal)

        configuration_id = proto.configuration_id

        package = GatewayPackage(
            configuration_id, signals, metadata
        )
        return package



#!/usr/bin/python

PYTHON   = 'Python'
PYSERIAL = 'PySerial'
PYWIN32  = 'PyWin32'
WEBPY    = 'WebPy'

REQUIRED_PYTHON_MAJOR = 2
REQUIRED_PYTHON_MINOR = 6

#======================== private =========================================

def _verifyPython():
    import sys
    import platform

    output = []
    
    # verify Python version running -- use the tuple format for compatibility,
    # .major, .minor, etc are only present in 2.7
    major   = sys.version_info[0]
    minor   = sys.version_info[1]
    micro   = sys.version_info[2]
    output += [
        "You are running Python {0}.{1}.{2} on platform {3}, {4}".format(
            major,
            minor,
            micro,
            platform.platform(),
            platform.machine()
        )
    ]
    
    # is that enough?
    if major==REQUIRED_PYTHON_MAJOR and minor>=REQUIRED_PYTHON_MINOR:
        goodToGo = True
    else:
        output += ["You need at least Python {0}.{1}".format(REQUIRED_PYTHON_MAJOR,REQUIRED_PYTHON_MINOR)]
        goodToGo = False
        
    return (goodToGo,'\n'.join(output))

def _verifyPyserial():
    output = []
    pyserial_needed = False
    
    try:
        import serial
        goodToGo = True
        output +=  ["You have PySerial {0}".format(serial.VERSION)]
    except ImportError:
        goodToGo = False
        pyserial_needed = True

    if pyserial_needed:
        output +=  ["You need to install PySerial:"]
        output +=  [" - information http://pyserial.sourceforge.net/"]
        try:
            import easy_install
            output +=  [" - install with 'easy_install pyserial'"]
        except:
            output +=  [" - download pyserial-2.5.win32.exe from http://sourceforge.net/projects/pyserial/files/pyserial/2.5/"]
        
    return (goodToGo,'\n'.join(output))

def _verifyPywin32():
    import os

    if not os.name in ['nt']:
        return (True, 'PyWin32 is not used on non-Windows systems.')
    
    output = []
    pywin32_needed = False

    try:
        import pywintypes
        import win32api
        goodToGo = True
        try:
            fixed_file_info = win32api.GetFileVersionInfo(win32api.__file__, '\\')
            pywin32_ver = fixed_file_info['FileVersionLS'] >> 16
            output += ["You have PyWin32 build {0}".format(pywin32_ver)]
        except pywintypes.error:
            # pypiwin32 can not detect the version resource in its own win32api file
            output += ["PyWin32 is installed via pypiwin32, but pypiwin32 cannot find the version"]
    except ImportError:
        goodToGo = False
        pywin32_needed = True

    if pywin32_needed:
        output +=  ["You need to install PyWin32:"]
        output +=  [" - information http://sourceforge.net/projects/pywin32/"]
        output +=  [" - download and install the latest release for your Python version from http://sourceforge.net/projects/pywin32/files/pywin32/"]
        
    return (goodToGo,'\n'.join(output))

def _verifyWebPy():
    output = []
    webpy_needed = False
    
    try:
        import web
        goodToGo = True
        output +=  ["You have web.py intalled"]
    except ImportError:
        goodToGo = False
        webpy_needed = True

    if webpy_needed:
        output +=  ["You need to install web.py:"]
        output +=  [" - information http://webpy.org/"]
        try:
            import easy_install
            output +=  [" - install with 'easy_install web'"]
        except:
            output +=  [" - follow installation procedure at http://webpy.org/install"]
        
    return (goodToGo,'\n'.join(output))


ComponentTests = { PYTHON:   _verifyPython,
                   PYSERIAL: _verifyPyserial,
                   PYWIN32:  _verifyPywin32,
                   WEBPY:    _verifyWebPy, }

#======================== public ==========================================

def verifyComponents(elementsToTest):
    for elem in elementsToTest:
        try:
            (goodToGo,reason) = ComponentTests[elem]()
        except KeyError:
            goodToGo = False
            reason = 'error: unknown component to verify: '+str(elem)

        if goodToGo==False:
            return (goodToGo,reason)

    # HACK: InstallTest calls this with one component at a time, so we return the
    # last (only) reason which contains a message about the current version.
    # Other components mostly care about the goodToGo result, and only look at
    # the reason on failure.
    return (True, reason)



"""
This module defines the arm controller to admin the arm as a shared physical resource

"""
# ============================ imports =========================================
import threading
import time
from typing import List, Callable

from txp import settings
import logging

from txp.common.edge import ImageSignal, ThermalImageSignal, EdgeDescriptor
from txp.devices.drivers.arm_robot.arm_handler import ArmHandler
from txp.devices.drivers.usb.camera.camera_handler_base import CameraHandlerBase, CameraType

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)

class ArmController:

    def __init__(self, edge_descriptor:EdgeDescriptor):
        log.info("ArmController is instanciated")
        self.descriptor = edge_descriptor
        self._arm_handler = ArmHandler()
        self._camera_handlers: List[CameraHandlerBase] = []
        self._set_camera_handlers(edge_descriptor)
        self._lock = threading.Lock()

    def execute_arm_action(self, angle):
        """
        TODO: thinking on the production to take only the angles of connected servo based on its status...
              then use Arm_serial_servo_write instead of Arm_serial_servo_write6

        :param angle: tuple of 6 angles for the servo.
        :return: Signal images asociated to the arm.one ImageSignal ThermalImageSignal

        """
        acquired = self._lock.acquire(blocking=False)
        if not acquired:
            log.debug(f"Arm resource is locked when tryin to move to {angle}. "
                      f"Return immediately.")
            return []
        else:
            log.info("acquiring arm resource by driver angle {}".format(angle))
            images = []
            try:
                servo_status = self._arm_handler.arm_connection_status()
                log.info("servo status is {}".format(servo_status))
                if servo_status:
                    log.info("moving angle {}".format(angle))
                    self._arm_handler.set_arm_angle_position(angle)
                    images = self._get_signals_image()

            except Exception as ex:
                log.error("Exception in ArmController {}".format(ex))
            finally:
                self._lock.release()
                log.info("images captured resource released")
                return images

    def disconnect(self):
        self._lock.acquire()
        try:
            for camera in self._camera_handlers:
                camera.disconnect()
            self._camera_handlers = []
            self._arm_handler.flush()
        finally:
            self._lock.release()
            log.info("disconnecting released resource")

    def is_arm_fully_connected(self):
        """
        Validates that all cameras and servos are connected and working.
        :return: True is all resources are connected False if any physical device is in failure
        """
        log.info("validating if it is fully connected")
        connected = True
        try:
            for camera in self._camera_handlers:
                if not bool(camera.is_connected()):
                    connected = False
                log.info("camera {} status is {}".format(camera.get_camera_id(),connected))
            if not self._arm_handler.arm_connection_status():
                connected = False
        except Exception as ex:
            log.error("Arm is not fully connected exception {}".format(ex) )
            connected = False
        return connected

    def _get_signals_image(self):
        """
        Returns signal images
        :return: returns the signals of the images
        """
        log.info("getting camera images")
        image_signals = []
        for camera in self._camera_handlers:
            if camera.is_connected():
                image = camera.get_image()
                camera_type = camera.get_camera_type()
                if camera.get_camera_type() == CameraType.NORMAL:
                    image_signals.append(ImageSignal.build_signal_from_lists(
                        [image],
                        (1, len(image)),
                        0,
                        0
                    ))
                else:
                    image_signals.append(ThermalImageSignal.build_signal_from_lists(
                        [image],
                        (1, len(image)),
                        0,
                        0
                    ))
                log.debug("getting image from {}".format(camera_type))
            else:
                log.error("camera {} is not connected for {}".format(camera_type,self.descriptor.logical_id))
                return None
        return image_signals

    def _set_camera_handlers(self, edge_descriptor):

        sensors = edge_descriptor.get_virtual_sensors()

        for sensor  in sensors:
            if sensor.camera_type == CameraType.NORMAL:
                camera_handler_base = CameraHandlerBase(str(CameraType.NORMAL), sensor.physical_id)
                camera_handler_base.connect()
                log.info("Created camera handler {} with physical ID {} and status is {}".format(CameraType.NORMAL,sensor.physical_id,camera_handler_base.is_connected()))
                self._camera_handlers.append(camera_handler_base)
            if sensor.camera_type == CameraType.THERMOGRAPHIC:
                camera_handler_base = CameraHandlerBase(str(CameraType.THERMOGRAPHIC), sensor.physical_id)
                camera_handler_base.connect()
                log.info("Created camera handler {} with physical ID {} and status is {}".format(CameraType.THERMOGRAPHIC,sensor.physical_id,camera_handler_base.is_connected()))
                self._camera_handlers.append(camera_handler_base)






from typing import List

import ray
import logging
import json

from pydantic import BaseModel, Field
from txp.cloud.common import utils as cu
from google.cloud import firestore
from google.oauth2 import service_account
from google.cloud import bigquery
from txp.cloud.controllers import bigquery as bq
from google.cloud import storage
import datetime
from txp.cloud import settings


class DatasetCommitPayload(BaseModel):
    """Expected Payload to receive when the client request the
    creation of """
    tenant_id: str
    task_id: str
    start: str
    end: str
    dataset_name: str
    dataset_versions: List[str]
    machine_id: str = Field(alias='asset_id')
    bigquery_dataset: str

    class Config:
        # https://github.com/samuelcolvin/pydantic/issues/602#issuecomment-503189368
        allow_population_by_field_name = True


@ray.remote
class DatasetCommitter:
    """This actor supports the datasets generation based on labeled data
    in TXP data sources.

    """

    def __init__(self, credentials_str: str, data_warehouse_bucket_name, log_level=logging.INFO):
        logging.basicConfig(level=log_level)

        self.data_warehouse_bucket_name = data_warehouse_bucket_name
        json_dict_service_account = json.loads(credentials_str, strict=False)
        self._credentials = service_account.Credentials.from_service_account_info(
            json_dict_service_account
        )

        logging.info(f"{self.__class__.__name__} Successfully created")

    def _get_dataset(
            self,
            tenant_id,
            table_name,
            edges,
            dataset_versions,
            start_datetime,
            end_datetime,
            bigquery_dataset,
            bigquery_client,
    ):
        return bq.get_all_signals_for_asset(
            tenant_id,
            f"{bigquery_dataset}.{table_name}",
            edges,
            dataset_versions,
            start_datetime,
            end_datetime,
            bigquery_client,
        )

    def commit_dataset(self, commit: DatasetCommitPayload):
        logging.info(
            f"Requested to commit Dataset with the following parameters: "
            f"tenant-id[{commit.tenant_id}]|"
            f"machine-id[{commit.machine_id}]|"
            f"task-id[{commit.task_id}]"
        )
        firestore_db: firestore.Client = firestore.Client(
            credentials=self._credentials, project=self._credentials.project_id
        )
        bigquery_client: bigquery.Client = bigquery.Client(
            credentials=self._credentials, project=self._credentials.project_id
        )
        machine = cu.get_machine_from_firestore(
            firestore_db, commit.tenant_id, commit.machine_id
        )
        if machine is None:
            logging.error(
                f'Tenant:{commit.tenant_id} or Machine:{commit.machine_id} not found'
            )
            return

        if commit.task_id not in machine["tasks"]:
            logging.error(
                f"No task found or task is not unique for {commit.task_id}"
            )
            return

        task = machine["tasks"][commit.task_id]

        logging.info(f'Machine {commit.machine_id} and task {commit.task_id} found '
                     f'successfully')

        edges = task["edges"]
        start_datetime = datetime.datetime.strptime(commit.start, settings.time.start_date_timezone_format)
        end_datetime = datetime.datetime.strptime(commit.end, settings.time.start_date_timezone_format)

        time_dataset = self._get_dataset(
            commit.tenant_id,
            "time",
            edges,
            commit.dataset_versions,
            start_datetime,
            end_datetime,
            commit.bigquery_dataset,
            bigquery_client,
        )
        fft_dataset = self._get_dataset(
            commit.tenant_id,
            "fft",
            edges,
            commit.dataset_versions,
            start_datetime,
            end_datetime,
            commit.bigquery_dataset,
            bigquery_client,
        )
        psd_dataset = self._get_dataset(
            commit.tenant_id,
            "psd",
            edges,
            commit.dataset_versions,
            start_datetime,
            end_datetime,
            commit.bigquery_dataset,
            bigquery_client,
        )
        time_metrics_dataset = self._get_dataset(
            commit.tenant_id,
            "time_metrics",
            edges,
            commit.dataset_versions,
            start_datetime,
            end_datetime,
            commit.bigquery_dataset,
            bigquery_client,
        )
        psd_metrics_dataset = self._get_dataset(
            commit.tenant_id,
            "psd_metrics",
            edges,
            commit.dataset_versions,
            start_datetime,
            end_datetime,
            commit.bigquery_dataset,
            bigquery_client,
        )
        fft_metrics_dataset = self._get_dataset(
            commit.tenant_id,
            "fft_metrics",
            edges,
            commit.dataset_versions,
            start_datetime,
            end_datetime,
            commit.bigquery_dataset,
            bigquery_client,
        )

        storage_client = storage.Client(credentials=self._credentials,
                                        project=self._credentials.project_id)

        bucket = storage_client.bucket(self.data_warehouse_bucket_name, user_project=self._credentials.project_id)

        task_path = cu.get_gcs_task_bucket_path(
            commit.tenant_id, commit.machine_id, commit.task_id
        )
        if len(time_dataset):
            blob = bucket.blob(
                cu.get_gcs_dataset_file_name(
                    task_path,
                    commit.dataset_name,
                    "time",
                    commit.dataset_versions,
                )
            )
            time_dataset.to_parquet("/tmp/time.gzip", compression="gzip")
            blob.upload_from_filename("/tmp/time.gzip")
            logging.info(f"Successfully created time dataset for {commit}")
        else:
            logging.warning(f"Nothing found at time table in commit {commit}")

        if len(fft_dataset):
            blob = bucket.blob(
                cu.get_gcs_dataset_file_name(
                    task_path, commit.dataset_name, "fft", commit.dataset_versions
                )
            )
            fft_dataset.to_parquet("/tmp/fft.gzip", compression="gzip")
            blob.upload_from_filename("/tmp/fft.gzip")
            logging.info(f"Successfully created fft dataset for {commit}")
        else:
            logging.info(f"Nothing found at fft table in commit {commit}")

        if len(psd_dataset):
            blob = bucket.blob(
                cu.get_gcs_dataset_file_name(
                    task_path, commit.dataset_name, "psd", commit.dataset_versions
                )
            )
            psd_dataset.to_parquet("/tmp/psd.gzip", compression="gzip")
            blob.upload_from_filename("/tmp/psd.gzip")
            logging.info(f"Successfully created psd dataset for {commit}")
        else:
            logging.info(f"Nothing found at psd table {commit}")

        if len(time_metrics_dataset):
            blob = bucket.blob(
                cu.get_gcs_dataset_file_name(
                    task_path,
                    commit.dataset_name,
                    "time_metrics",
                    commit.dataset_versions,
                )
            )
            time_metrics_dataset.to_parquet(
                "/tmp/time_metrics.gzip", compression="gzip"
            )
            blob.upload_from_filename("/tmp/time_metrics.gzip")
        else:
            logging.info(f"Nothing found at time_metrics table {commit}")

        if len(fft_metrics_dataset):
            blob = bucket.blob(
                cu.get_gcs_dataset_file_name(
                    task_path,
                    commit.dataset_name,
                    "fft_metrics",
                    commit.dataset_versions,
                )
            )
            fft_metrics_dataset.to_parquet("/tmp/fft_metrics.gzip", compression="gzip")
            blob.upload_from_filename("/tmp/fft_metrics.gzip")
            logging.info(f"Successfully created fft_metrics dataset for {commit}")
        else:
            logging.info(f"Nothing found at fft_metrics table {commit}")

        if len(psd_metrics_dataset):
            blob = bucket.blob(
                cu.get_gcs_dataset_file_name(
                    task_path,
                    commit.dataset_name,
                    "psd_metrics",
                    commit.dataset_versions,
                )
            )
            psd_metrics_dataset.to_parquet("/tmp/psd_metrics.gzip", compression="gzip")
            blob.upload_from_filename("/tmp/psd_metrics.gzip")
            logging.info(f"Successfully created psd_metrics dataset for {commit}")
        else:
            logging.info(f"Nothing found at psd_metrics table {commit}")

        storage_client.close()
        firestore_db.close()
        bigquery_client.close()



import ray
import logging
import json
from google.oauth2 import service_account
import google.cloud.firestore as firestore
from txp.ml.training_service.model_trainers.factory import TrainerFactory
from txp.cloud.common import utils as cu


@ray.remote
class TrainingManager:
    """This class is in charge of processing anything that arrives at the
    /training endpoint.

    This class will be called from the main Ray Serve deployment serving
    the HTTP endpoints for training, and it will:
        - create new Actors to perform training related activities
        - interact with in memory actors refs
    """

    def __init__(self, credentials_str, models_bucket_name, data_warehouse_bucket_name,
                 prediction_service_topic, models_registry_collection_name, log_level=logging.INFO):
        self._log_level = log_level
        logging.basicConfig(level=log_level)
        json_dict_service_account = json.loads(credentials_str, strict=False)
        credentials = service_account.Credentials.from_service_account_info(
            json_dict_service_account
        )
        self._credentials_str = credentials_str
        self._firestore_db = firestore.Client(
            credentials=credentials, project=credentials.project_id
        )
        self.factory = TrainerFactory(credentials_str,  models_bucket_name, data_warehouse_bucket_name,
                                      prediction_service_topic, models_registry_collection_name, log_level)
        logging.info(f"{self.__class__.__name__} Created successfully.")

    def process_acknowledge(self, acknowledge) -> bool:
        task_definition = cu.get_task_from_firestore(
            self._firestore_db,
            acknowledge["tenant_id"],
            acknowledge["machine_id"],
            acknowledge["task_id"],
        )
        if task_definition is None:
            logging.error(f"No Task JSON found for {acknowledge['task_id']}")
            return False

        task, trainer = self.factory.get_task_and_trainer(task_definition)
        trainer.train.remote(
            task,
            task_definition,
            acknowledge["dataset_name"],
            acknowledge["dataset_versions"],
            acknowledge["tenant_id"],
            acknowledge["machine_id"],
            acknowledge["task_id"],
        )
        logging.info(
            f"Successfully started training in cluster for"
            f"{acknowledge['task_id']}"
        )
        return True

    def process_publish(self, publish_command):
        task_definition = cu.get_task_from_firestore(
            self._firestore_db,
            publish_command["tenant_id"],
            publish_command["machine_id"],
            publish_command["task_id"],
        )
        if task_definition is None:
            logging.error(f"No Task JSON found for {publish_command['task_id']}")
            return False

        task, trainer = self.factory.get_task_and_trainer(task_definition)
        trainer.publish_model.remote(
            publish_command["tenant_id"],
            publish_command["machine_id"],
            publish_command["task_id"],
            publish_command["dataset_name"],
            publish_command["dataset_versions"],
        )
        logging.info(
            f"Successfully started training in cluster for {publish_command['task_id']}"
        )
        return True



"""
This module implements the Telemetry Sender object.

The telemetry sender object is responsible for processing the
dispatch of Gateway Packages to the cloud.
"""

# ============================ imports =========================================
from txp.devices.package_queue import PackageQueueProxy
from txp.devices.package import GatewayPackage
from txp.common.configuration import GatewayConfig
from txp.devices.mqtt import GatewayMQTTClient, MQTTMessageType, MQTTSentPayloadContext
from txp.devices.tasks_scheduler import ScheduleTask, TasksScheduler
from typing import Callable, Dict, List, Set, Optional, Union
from txp.devices.exceptions import *
import datetime
import threading
import logging
import time
import base64
import queue
import enum
from txp.common.config import settings

log = logging.getLogger(__name__)

log.setLevel(settings.txp.general_log_level)


# ==============================================================================
# Telemetry Sender class
# ==============================================================================
class OverlappedDispatchError(Exception):
    """Raised when the TelemetrySender is requested to start a dispatch processing, and there's already
    one being executed."""

    pass


class NotDispatchProcessingStarted(Exception):
    """Raised when the TelemetrySender is requested to stop a dispatch processing, and there's already
    one being executed"""

    pass


# ==============================================================================
# _PackagesBuffer class.
# ==============================================================================
class _PackagesBuffer:
    """This class abstracts a buffer used in the telemetry sending logic.

    This buffer allows to store packages per device, and to ask/return packages
    per device.

    It also supports to free itself when the buffer is overflow.
    When this happens, the buffer will return all the packages that contains
    in a sequential interspersed way.

    The buffer will be overflowed at some point, but it'll accept new packages.
    The client class is responsible for detecting the overflow and apply the
    appropriate logic.
    """

    def __init__(
        self,
        max_num_packages: int,
        edges_serials: List[str],
    ):
        """Args:

        max_num_packages: The maximum number of packages before overflow.
        edges_serials: The list of serials of edge devices that will store packages
        """
        self._max_num_packages: int = max_num_packages
        self._curr_num_packages: int = 0
        self._device_packages_table: Dict[str, queue.Queue] = {}
        self._build_device_packages_table(edges_serials)

        self._active_holds_serials: Set[str] = set()
        """_active_holds_serials: a set to know in constant time which serials has packages in
        in the table"""

    def _build_device_packages_table(self, edges_serials: List[str]):
        self._device_packages_table.clear()
        for serial in edges_serials:
            log.debug(f"Serial registered in buffer table: {serial}")
            self._device_packages_table[serial] = queue.Queue()

    def has_package(self, serial: str) -> bool:
        """Return True if the buffer contains a package for the device serial."""
        if serial not in self._device_packages_table:
            raise RuntimeError(f"Edge Serial not found in buffer: {serial}")

        return self._device_packages_table[serial].empty()

    def get_package(self, serial: str) -> GatewayPackage:
        """Returns the most recent package for the given device serial.

        Raises:
            queue.Empty if there's no package to retrieve. Call the has_package()
            method to ensure a safe call.

        Returns:
            GatewayPackage object.
        """
        package = self._device_packages_table[serial].get(block=False)
        self._curr_num_packages = self._curr_num_packages - 1
        if self._device_packages_table[serial].empty():
            self._active_holds_serials.remove(package.metadata.edge_descriptor.logical_id)
        return package

    def overflow(self) -> bool:
        """Returns True if the Gateway reached its max capacity."""
        return self._curr_num_packages >= self._max_num_packages

    def has_packages(self) -> bool:
        """Returns True if the buffer holds 1 or more packages"""
        return bool(self._active_holds_serials)

    def has_valid_packages_from_serials(self, serials: Set[str]) -> bool:
        """Returns True if there is at least one package from one of the
        serials list in the buffer."""
        intersect = self._active_holds_serials.intersection(serials)
        return bool(intersect)

    def get_valid_package_from_serials(self, serials: Set[str]) -> GatewayPackage:
        serial = self._active_holds_serials.intersection(serials).pop()
        return self.get_package(serial)

    def insert_package(self, package: GatewayPackage) -> None:
        """Inserts a package in the Buffer.

        Raises:
            KeyError if the logical_id is not found in the internal table
            of the buffer. This case should never happen.

        Returns:
            None
        """
        logical_id = package.metadata.edge_descriptor.logical_id
        log.debug(f"Package inserted in buffer. Device Serial: {logical_id}")
        self._device_packages_table[logical_id].put(package)
        self._curr_num_packages = self._curr_num_packages + 1
        self._active_holds_serials.add(logical_id)

    def free_buffer(self) -> List[GatewayPackage]:
        """Returns all the packages hold by the buffer and leaves the overflow
        state.

        The packages will be returned in a list in a sequential interspersed way.

        Returns:
            List of GatewayPackage instances
        """
        packages = []
        while self._active_holds_serials:
            for serial in self._active_holds_serials.copy():
                packages.append(self.get_package(serial))

        return packages


# ==============================================================================
# Telemetry Rollover policies
# ==============================================================================
class TelemetryRolloverPolicy(enum.Enum):
    DISCARD = "discard"


# ==============================================================================
# Telemetry Sender class
# ==============================================================================
class TelemetrySender:
    """The TelemetrySender object is designed to deal with all the
    GatewayPackages dispatching activities.

    This object should:
        - Read GatewayPackages from the gateway packages queue.
        - Send the packages in a manner that provides a fair share of
            the connection.
        - Interpret input information that allows it to understand how
            to schedule telemetry sending operations.

    The TelemetrySender scheduling is specified in the configuration of the Job
    currently set in the Gateway. Currently the configuration of the TelemetrySender
    requires:
        - start_date: A date to start the Telemetry Sender.
        - duration: The duration of the telemetry dispatch task.
        - frequency: The frequency to repeat the dispatch.
    """

    def __init__(
            self,
            gateway_packages_queue: PackageQueueProxy,
            gateway_mqtt_client: GatewayMQTTClient,
            gateway_process_stats_callback: Callable,
            telemetry_task_config: Optional[Dict],
            gateway_config: GatewayConfig,
            edge_serials: List[str]
    ):
        """
        Args:
            gateway_packages_queue: The Gateway centralized and shared packages
                Queue.
            gateway_mqtt_client: The GatewayMQTTClient instance used by the Gateway.
            gateway_process_stats_callback: A callback provided by the Gateway to process
                sent packages stats.
            telemetry_task_config: The telemetry task configuration described in the class
                docstring. It can be None at set later with instance methods.
            gateway_process_stats_callback: A callback provided by the Gateway to process
                sent packages stats.
            edge_serials: A list with the edge devices serials monitored by the gateway.
        """
        self._rollover_policy = self._parse_rollover_policy(gateway_config.job.telemetry_rollover_policy)
        log.info(f"Telemetry Sender set rollover policy to: {self._rollover_policy}")

        self._packages_queue: PackageQueueProxy = gateway_packages_queue
        self._mqtt_client: GatewayMQTTClient = gateway_mqtt_client
        self._process_stats: Callable = gateway_process_stats_callback

        self._send_events_thread: threading.Thread = None
        self._send_events_stop: threading.Event = threading.Event()

        # Fair dispatch attributes
        self._serials_weights_table: Dict[str, int] = {}
        """_serials_weights_table: a table of the allowed number of packages per fair segment
        for edge."""
        self._build_serials_weights_table(gateway_config, edge_serials)

        self._sent_packages_table: Dict[str, int] = {}
        """_sent_packages_table: a table to be used when sending packages in fair segments, to hold
        the count of sent packages per serial."""
        self._restart_sent_packages_table()

        self._valid_serials_to_send: Set[str] = set()
        """_valid_serials_to_send: a set of serials that are valid when building a fair segment."""

        # TODO: the max number for the buffer is the number of edges. This should be refined.
        self._buffer: _PackagesBuffer = _PackagesBuffer(len(edge_serials), edge_serials)

        self._unsent_packages_registry: queue.Queue[Union[bytes, str, MQTTSentPayloadContext]] = queue.Queue()

        """_unsent_packages_registry: internal table to hold the unsent packages. This table allows 
        to retry sending unacknowledged messages based on the rollover policies."""
        self._unsent_packages_check_interval: int = settings.gateway.telemetry.unsent_packages_check_interval

    # ============================== Private Methods ================================================
    @staticmethod
    def _parse_rollover_policy(value: str) -> TelemetryRolloverPolicy:
        try:
            policy = TelemetryRolloverPolicy(value)
            return policy
        except:
            log.error("Unknown excpetion while parsing Telemetry Rollover Policy. Set to default Discard.")
            return TelemetryRolloverPolicy.DISCARD

    def _build_serials_weights_table(
            self, gateway_config: GatewayConfig, edge_serials: List[str]
    ):
        """Builds the weights table for the packages.

        Note: Right now, treat all packages as equals, with the same weight of 1.
        TODO: Implement weights compute.

        Args:
            gateway_config: The Gateway Configuration instance.
        """
        for serial in edge_serials:
            self._serials_weights_table[serial] = 1

    def _restart_sent_packages_table(self):
        serials = self._serials_weights_table.keys()
        for serial in serials:
            self._sent_packages_table[serial] = 0

    def _max_packages_reached_per_weight(self):
        serials = self._serials_weights_table.keys()
        return all(
            list(
                map(
                    lambda serial: self._sent_packages_table[serial]
                                   == self._serials_weights_table[serial],
                    serials,
                )
            )
        )

    def _initialize_new_segment(self):
        """Initializes attributes to track the building of a new fair segment"""
        self._restart_sent_packages_table()
        self._valid_serials_to_send = set(self._sent_packages_table.keys())

    def _is_valid(self, package: GatewayPackage):
        """Returns True if sending a package with the serial is fair."""
        serial = package.metadata.edge_descriptor.logical_id
        return (
                self._sent_packages_table.setdefault(serial, 0)
                < self._serials_weights_table[serial]
        )

    def _send_event(self, package: GatewayPackage) -> None:
        """Sends the Package to the cloud."""
        edge_serial = package.metadata.edge_descriptor.logical_id
        telemetry_topic = settings.gateway.mqtt.telemetry_topic_batch_template.format(
            edge_serial
        )

        if settings.gateway.streaming_mode == "REALTIME":
            telemetry_topic = settings.gateway.mqtt.telemetry_topic_realtime_template.format(
                edge_serial
            )

        proto = package.get_proto()
        encoded_payload = base64.b64encode(proto.SerializeToString())
        sent = False
        try:
            if len(encoded_payload) > 250000:
                log.warning(
                    f"Package exceeds max valid size. Will not Publish. Package Size: {len(encoded_payload)}"
                )

            else:
                self._mqtt_client.publish_message(
                    MQTTMessageType.TELEMETRY_MESSAGE,
                    edge_serial,
                    telemetry_topic,
                    encoded_payload,
                    attach=True,
                    track_acknowledge=True,
                    try_reconnect=False
                )
                sent = True
                log.info(f"Sent Package from edge: {edge_serial}")

        except MqttMessageNotPublished:
            log.error(f"Error with disconnected MQTT client while publishing "
                      f"telemetry event.")
            self._unsent_packages_registry.put(
                MQTTSentPayloadContext(
                    edge_serial, telemetry_topic, encoded_payload, True
                )
            )

        except:
            log.error(f"Unknownn exception while publishing telemetry event. ")
            self._unsent_packages_registry.put(
                MQTTSentPayloadContext(
                    edge_serial, telemetry_topic, encoded_payload, True
                )
            )

        finally:
            self._process_stats(package, sent)

    def _check_unsent_events(self) -> List:
        """Checks if there are unsent messages in the physical layer, and clears the
        sent messages from the physical layer."""
        log.debug("Checking if there are unsent events...")
        unsent_messages = self._mqtt_client.get_unacknowledged_messages(MQTTMessageType.TELEMETRY_MESSAGE)
        if unsent_messages:
            log.info(f"Received {len(unsent_messages)} unsent messages to sent in rollover")
            for payload_context in unsent_messages:
                self._unsent_packages_registry.put(payload_context)

        self._mqtt_client.clear_acknowledged_messages(MQTTMessageType.TELEMETRY_MESSAGE)
        log.debug(f"Telemetry current number of unsent packages: {self._unsent_packages_registry.qsize()}")

    def _register_package_sent(self, package: GatewayPackage):
        """Register the sent package in the _sent_packages_table"""
        serial = package.metadata.edge_descriptor.logical_id
        self._sent_packages_table[serial] += 1
        if self._sent_packages_table[serial] == self._serials_weights_table[serial]:
            self._valid_serials_to_send.remove(serial)

    def _send_packages(self, timeout: int) -> None:
        """Private method for sending packages following an algorithm that tries to send
        fair sequences of packages.

        This is the callback to be called by the Tasks in the TaskScheduler, to send
        telemetry.

        Fair sequences are determined by the computed table of weights for the devices.

        This function will loop until the passed callback returns True. This enables to reuse
        accordingly from different places.

        Note: There's a limit of 60.000 messages per minute established by IoT Core.
            https://cloud.google.com/iot/quotas#rate_limits. This code will not surpass that limit.

        Args:
            timeout: the timeout for the sending process to finish.
        """
        total_sent_packages = 0
        sent_messages_limit_count = 0
        limit_per_minute = 60000  # Default 60000 messages per minute
        start_time = datetime.datetime.now()
        elapsed_time = start_time
        unsent_packages_check = start_time
        messages_limit_check = start_time

        log.info(f'Starting scheduled Telemetry sending task at: {start_time}')

        self._initialize_new_segment()

        if self._buffer.has_packages():
            log.warning("Lag: packages found in Buffer from previous run.")

        try:
            while True:
                # Check if Telemetry time is over
                if (elapsed_time - start_time).total_seconds() > timeout:
                    break

                elapsed_time = datetime.datetime.now()

                # Check if a minute has passed to refresh messages limit count
                if (elapsed_time - messages_limit_check).total_seconds() > 60:
                    sent_messages_limit_count = 0
                    messages_limit_check = datetime.datetime.now()

                # If not, then check if messages limit has been surpassed
                elif sent_messages_limit_count == limit_per_minute:
                    wait_time = (
                            (messages_limit_check + datetime.timedelta(minutes=1)) - messages_limit_check
                    ).total_seconds()
                    log.info(f"Published messages limit of 60.000 p/minute reached. Wait for {wait_time} secs")
                    time.sleep(wait_time)

                # Check for unacknowledged packages
                if ((elapsed_time - unsent_packages_check).total_seconds()
                        > self._unsent_packages_check_interval):
                    unsent_packages_check = datetime.datetime.now()
                    self._check_unsent_events()
                    log.debug(f"Telemetry current number of unsent packages: "
                              f"{self._unsent_packages_registry.qsize()}")

                # Check if somebody requested stop.
                if self._send_events_stop.is_set():
                    log.info(f"Dispatch packages completed. Packages sent: {total_sent_packages}")
                    log.info(f'Scheduled Dispatch task finished.')
                    self._discard_policy()
                    return

                # Here starts the sending algorithm
                if self._buffer.has_packages():
                    log.debug("Buffer has packages. Trying to get valid package")
                    if self._buffer.has_valid_packages_from_serials(
                        self._valid_serials_to_send
                    ):
                        log.debug("Buffer has a valid package. Sending package.")
                        package = self._buffer.get_valid_package_from_serials(
                            self._valid_serials_to_send
                        )
                        self._send_event(package)
                        self._register_package_sent(package)
                        total_sent_packages = total_sent_packages + 1  # this might count a fake positive
                        sent_messages_limit_count += 1
                        continue
                    else:
                        log.debug("Buffer has not valid package. Trying to get from Queue.")

                if self._packages_queue.size() == 0:
                    packages = self._buffer.free_buffer()
                    if packages:
                        log.info("No packages received in Queue. Sending the buffered packages anyways.")
                    for package in packages:
                        self._send_event(package)
                        total_sent_packages = total_sent_packages + 1  # this might count a fake positive
                        sent_messages_limit_count += 1
                    self._initialize_new_segment()
                    continue

                try:
                    package = self._packages_queue.get_package()
                except Exception as e:
                    log.error(f"Exception while getting package from Gateway Queue: {e}. ")
                    package = None

                # Tries to send unacknowledge messages
                if self._unsent_packages_registry.qsize():
                    message_payload = self._unsent_packages_registry.get()
                    log.info("Retry to publish unsent telemetry message")
                    self._retry_sent_payload(message_payload)

                if package is None:
                    log.debug(
                        f"None package received in Queue."
                    )
                    continue

                log.debug(
                    f"Package received in Queue from: {package.metadata.edge_descriptor.logical_id}"
                )
                if self._is_valid(package):
                    log.debug("The Package is valid. Sending package.")
                    self._send_event(package)
                    self._register_package_sent(package)
                    total_sent_packages = total_sent_packages + 1  # this might count a fake positive
                    sent_messages_limit_count += 1
                else:
                    log.debug(
                        "The package can not be sent in the current fair segment. Saving package in buffer."
                    )
                    self._buffer.insert_package(package)

                if self._buffer.overflow():
                    log.debug("Buffer has overflow. Sending packages in buffer.")
                    log.info("Buffer has overflow. Sending packages in buffer.")
                    packages = self._buffer.free_buffer()
                    for package in packages:
                        self._send_event(package)
                        total_sent_packages = total_sent_packages + 1  # this might count a fake positive
                        sent_messages_limit_count += 1
                    self._initialize_new_segment()
                    continue

                if self._max_packages_reached_per_weight():
                    log.debug("Fair segment sent. Next iteration will start to build a new fair segment.")
                    self._initialize_new_segment()

            # Check for pending packages
            log.info(f"Dispatch packages completed assigned time. "
                     f"Approximated packages sent: {total_sent_packages}")

            self._check_unsent_events()
            if (
                self._unsent_packages_registry.qsize() > 0 or
                self._packages_queue.size() > 0
            ):
                log.info("There are packages without sending after Telemetry configured time."
                         f"Unsent packages: {self._unsent_packages_registry.qsize()}. "
                         f"Packages in Queue: {self._packages_queue.size()}")
                self._apply_rollover_policy()

        except Exception as e:
            # The _send_packages call will return and the new telemetry task will be scheduled.
            log.error(f'Unexpected exception while sending packages: {e}')

    def _retry_sent_payload(self, payload_context: MQTTSentPayloadContext):
        try:
            self._mqtt_client.publish_message(
                MQTTMessageType.TELEMETRY_MESSAGE,
                payload_context.device_id,
                payload_context.topic,
                payload_context.payload,
                attach=True,
                track_acknowledge=True,
                try_reconnect=False
            )
            sent = True
            log.info(f"Sent Package from edge: {payload_context.device_id}")

        except MqttMessageNotPublished:
            log.error(f"Error with disconnected MQTT client while publishing "
                      f"telemetry event.")
            self._unsent_packages_registry.put(
                MQTTSentPayloadContext(
                    payload_context.device_id, payload_context.topic, payload_context.payload, True
                )
            )

        except:
            log.error(f"Unknownn exception while publishing telemetry event. ")
            self._unsent_packages_registry.put(
                MQTTSentPayloadContext(
                    payload_context.device_id, payload_context.topic, payload_context.payload, True
                )
            )

    def _apply_rollover_policy(self):
        """
        TODO: implement the other policies for Telemetry rollover:
            - next_x_windows: This policy allows to send pending telemetry packages in
                the next X observation times.
        """
        if self._rollover_policy == TelemetryRolloverPolicy.DISCARD:
            self._discard_policy()

    def _discard_policy(self):
        log.info("Telemetry rollover policy 'Discard' will be executed")
        self._unsent_packages_registry = queue.Queue()
        self._packages_queue.clear()
        self._buffer.free_buffer()
        self._initialize_new_segment()  # resets buffer internal tables

    def _get_scheduled_task(self, telemetry_task_config: Dict) -> ScheduleTask:
        """Get a List of scheduled tasks to configure in the TasksScheduler.

        TODO: The task scheduled for telemetry dispatch should be computed based on the Job
            sampling window. This is just filling the required data fields.

        Args:
            telemetry_task_config: A dictionary with the telemetry configuration for the concept of "Job".
        """
        task_end_timeout = telemetry_task_config['duration'] * 60

        start_time = datetime.datetime.strptime(telemetry_task_config['start_time'],
                                                settings.gateway.tasks_scheduler.time_format)

        week_days_list = list(map(lambda char: int(char), telemetry_task_config['active_week_days']))

        # TODO:Telemetry sending end date should be computed based on the Job end date.
        #   Support this when supporting the full JOB configuration.
        required_fixed_end_date = datetime.datetime.strptime(
            settings.gateway.tasks_scheduler.max_date_string,
            settings.gateway.tasks_scheduler.start_date_format
        )

        task = ScheduleTask(
            datetime.datetime.now().date(), required_fixed_end_date.date(),
            start_time.time(), start_time.time(),
            week_days_list, self._send_packages, (task_end_timeout,)
        )
        return task

    # ============================== Public Methods ================================================
    def start(self, gateway_conf: GatewayConfig, devices: List[str], timeout: int):
        self._serials_weights_table.clear()
        self._build_serials_weights_table(gateway_conf, devices)
        self._sent_packages_table.clear()
        self._restart_sent_packages_table()
        self._valid_serials_to_send.clear()
        self._buffer = _PackagesBuffer(len(devices), devices)
        self._send_events_thread = threading.Thread(
            target=self._send_packages,
            name="telemetry-sending",
            args=(timeout,)
        )
        self._send_events_thread.start()
        self._send_events_stop.clear()

    def stop(self):
        self._send_events_stop.set()
        while self._send_events_thread.is_alive():
            time.sleep(1)

    def is_running(self) -> bool:
        """Returns True if the TelemetrySender is running some tasks scheduling
        in background"""
        return self._send_events_thread is not None and self._send_events_thread.is_alive()



import logging


class StateController:
    """ This class could confuse a little take your time reading the code.
    This class is in charge of saying which generated state is going to be processed.
    Into a Gateway Task we have multiple sampling windows for a given asset, this sampling windows
    cloud arrive to the prediction service in a random order, this class gives an order to these
    states.
    An ideal behavior should be to process sampling windows in ascending order, HOWEVER sampling windows could get
    lost during the telemetry process, this class tries to solve this problem with an slicing windows approach, this
    is:
    lets define a sequence of sampling windows as S1, S2, S3, ...., Sn, Sn+1, ...., suppose some sampling windows
    are lost somewhere and do not arrived to the service: S1, S2, X, X, X, X, S7, ..., here the X represents lost
    sampling windows, this class defines a windows tolerance, if the number of lost sampling windows is bigger than
    SAMPLING_WINDOWS_TOLERANCE then the lowest non processed sampling window is going to be processed, in the example
    shown above, suppose S1 was the last processed sampling window and S7 is the last arrived sampling window to the
    state manager, then the slicing window looks like:
    S1, S2, X, X, X, X, S7, ...,
        [l              r]
    since r - l + 1 >= SAMPLING_WINDOWS_TOLERANCE, the next sampling window to be processed is S2.

    This class also implements an states_history this is used for checking the history of states in order to decide
    if the state that is going to be processed is going to be stored in BQ and generate a notification or not. The
    size of this cache is given by MAX_NUMBER_OF_ENQUEUE_STATES

    """

    MAX_NUMBER_OF_ENQUEUE_STATES = 10
    SAMPLING_WINDOWS_TOLERANCE = 3

    def __init__(self, log_level=logging.INFO):
        logging.basicConfig(level=log_level)
        self.states_db = {}
        self.states_history = {}

    def save_state(self, state, gateway_task_id_, sampling_window_index, number_of_sampling_windows):
        logging.info(f"Storing state on StateController {state}")
        sampling_window_index_str = str(sampling_window_index)
        gateway_task_id = str(gateway_task_id_)
        if gateway_task_id in self.states_db:
            if int(sampling_window_index_str) >= self.states_db[gateway_task_id]["next"]:
                logging.info(f"Storing state on StateController db {state}")
                self.states_db[gateway_task_id]["states"][sampling_window_index_str] = state
            else:
                logging.info(
                    f'State not stored next expected state is {self.states_db[gateway_task_id]["next"]} however '
                    f'{int(sampling_window_index_str)} found. It means an already discarded state appeared. {state}')
        else:
            logging.info(f"Storing state on StateController db {state}, this is the first state received for "
                         f"{gateway_task_id}")
            self.states_db[gateway_task_id] = {
                "next": 0,
                "number_of_sampling_windows": number_of_sampling_windows,
                "states": {sampling_window_index_str: state}
            }

    def get_state(self, gateway_task_id_):

        logging.info(f"Getting new state to be processed for {gateway_task_id_}")

        def log_state_processed(observation_timestamp, index, total):
            logging.info(f'New state is going to be processed, '
                         f'observation timestamp: {observation_timestamp}, '
                         f'sampling windows index: {index}, '
                         f'total number of sampling windows: {total}')

        gateway_task_id = str(gateway_task_id_)
        if gateway_task_id not in self.states_db:
            return None
        next_index = str(self.states_db[gateway_task_id]["next"])

        if next_index in self.states_db[gateway_task_id]["states"]:
            state = self.states_db[gateway_task_id]["states"].pop(next_index)
            log_state_processed(state["observation_timestamp"], next_index,
                                self.states_db[gateway_task_id]["number_of_sampling_windows"])
            self.states_db[gateway_task_id]["next"] += 1
            if self.states_db[gateway_task_id]["next"] == self.states_db[gateway_task_id]["number_of_sampling_windows"]:
                self.states_db.pop(gateway_task_id)
            return state
        else:
            logging.info(f"Expected state with index {next_index} is not available yet. Looking for window size for "
                         f"{gateway_task_id}")
            if self.states_db[gateway_task_id]["states"]:
                upper_index = max(self.states_db[gateway_task_id]["states"], key=lambda k: int(k))
                lower_index = min(self.states_db[gateway_task_id]["states"], key=lambda k: int(k))
                if (int(upper_index) - int(next_index)) >= StateController.SAMPLING_WINDOWS_TOLERANCE:
                    logging.info(f"Current current window size for {gateway_task_id} is big enough, processing oldest "
                                 f"available state")
                    state = self.states_db[gateway_task_id]["states"].pop(lower_index)
                    self.states_db[gateway_task_id]["next"] = int(lower_index) + 1
                    log_state_processed(state["observation_timestamp"], lower_index,
                                        self.states_db[gateway_task_id]["number_of_sampling_windows"])
                    return state
            logging.info(f"Window size is not big enough for {gateway_task_id}, no state is going to be processed")
            return None

    def enqueue_state(self, state, gateway_task_id_):
        gateway_task_id = str(gateway_task_id_)
        if gateway_task_id not in self.states_history:
            self.states_history[gateway_task_id] = [state]
        else:
            self.states_history[gateway_task_id] = [state] + self.states_history[gateway_task_id]
        if len(self.states_history[gateway_task_id]) > StateController.MAX_NUMBER_OF_ENQUEUE_STATES:
            logging.info(f"Max number of states reached for {gateway_task_id}, popping oldest state")
            self.states_history[gateway_task_id].pop()

    def get_states_history(self, gateway_task_id_, number_of_states):
        gateway_task_id = str(gateway_task_id_)
        if gateway_task_id in self.states_history:
            return self.states_history[gateway_task_id][:number_of_states]
        else:
            return []

    def clear_history(self, gateway_task_id_):
        gateway_task_id = str(gateway_task_id_)
        if gateway_task_id not in self.states_db:
            logging.info(f"All states processed for {gateway_task_id}, deleting its history")
            self.states_history.pop(gateway_task_id)



from google.cloud import bigquery
from google.oauth2 import service_account
import json
import sys

CREDENTIALS_PATH = "../../../txp_cloud/pipelines/credentials/pub_sub_to_bigquery_credentials.json"
credentials = service_account.Credentials.from_service_account_file(CREDENTIALS_PATH)
client = bigquery.Client(credentials=credentials, project=credentials.project_id)


def create_table(table, dataset):
    project = credentials.project_id
    table_id = f"{project}.{dataset}.{table}"
    f = open(f'{table}.json')
    schema = json.load(f)
    client.delete_table(table_id, not_found_ok=True)
    table = bigquery.Table(table_id, schema=schema)
    table.time_partitioning = bigquery.TimePartitioning(
        type_=bigquery.TimePartitioningType.HOUR,
        field="partition_timestamp",
        expiration_ms=13824000000
    )
    table.clustering_fields = ["tenant_id", "asset_id"]
    client.create_table(table)
    f.close()


def main():
    args = sys.argv[1:]
    if not len(args):
        print("Dataset not specified")
        return
    dataset = args[0]
    create_table("events", dataset)
    create_table("states", dataset)


if __name__ == "__main__":
    main()



"""
This module exports a function that can be deployed as a GCP Cloud function
capable of updating the TXP Firestore configuration based on the manual
configuration editable collection.

The Manual Configuration Editable collection is a special collection in our
Firestore database, that can be used to edit configurations using the Firestore
web console.

This is provided as a way to edit configuration while we develop a web client
to communicate with Firestore.

TODO: This script should disappear eventually. We should have a database initialization
    script stored somewhere.
"""

# =========================== Imports ===============================
import google.cloud.firestore as firestore
import txp.cloud.controllers.firestore as txp_firestore_controller
from txp.cloud import (
    pull_current_configuration_from_firestore
)
from txp.cloud import settings
import txp.common.configuration.config_version as txp_config_version
from typing import List, Dict, Union
import pytz
import datetime
import logging

log = logging.getLogger(__name__)

# =========================== Constant Data ===============================
manual_configuration_collection = (
    settings.firestore.manual_configuration_edit_collection
)
config_form_document = settings.firestore.config_form_document
machines_group_collection_name = settings.firestore.machines_group_collection
machines_collection_name = settings.firestore.machines_collection
gateways_collection_name = settings.firestore.gateways_collection
devices_collection_name = settings.firestore.devices_collection
edges_collection_name = settings.firestore.edges_collection
jobs_collections_name = settings.firestore.jobs_collection
configurations_collection_name = settings.firestore.configurations_collection
zoned_datetime_str_format = settings.firestore.datetime_zoned_format
tenants_collection = settings.firestore.tenants_collection


# =========================== Helper Functions ===============================
def pull_docs_from_manual_form_collection(
        db: firestore.Client, configuration_document: str, subcollection: str, as_dict: bool = True
) -> Union[List[Dict], List[firestore.DocumentSnapshot]]:
    """Returns the Dict instances from the documents found in the
    collection for the manual configuration form collection"""
    documents_coll = (
        db.collection(manual_configuration_collection)
            .document(configuration_document)
            .collection(subcollection)
            .get()
    )

    documents_coll = list(filter(
        lambda doc: "template" not in doc.id, documents_coll
    ))

    if as_dict:
        documents_coll = list(map(
            lambda doc: doc.to_dict(), documents_coll
        ))

    return documents_coll


def _create_new_configuration(
        db: firestore.Client,
        tenant_reference: firestore.DocumentReference
) -> firestore.DocumentReference:
    """Creates a new entity (document) in the configurations root level collection.

    The configuration time is aware time for UTC zone.
    The configuration entity has a timestamp based on the Firestore server,
        in order to support queries by timestamp.

    Returns:
        The DocumentReference to the new configuration entity.
    """
    previous_configuration = pull_current_configuration_from_firestore(db, tenant_reference.get().get('tenant_id'))

    if (previous_configuration and
            previous_configuration.exists and
            'configuration_id' in previous_configuration.to_dict()):
        new_configuration_id = txp_config_version.get_next_normal_version(
            previous_configuration.to_dict()['configuration_id']
        )
    else:
        new_configuration_id = 1

    config_timestamp = datetime.datetime.now(pytz.timezone(settings.gateway.timezone)).replace(microsecond=0)
    ref = db.collection(configurations_collection_name).add(
        {
            "configuration_id": str(new_configuration_id),  # string to allow possible id generation in the future
            # TODO: Remove `since` key if nobody is using it. This is local time eventually turned into UTC by Firestore
            "since": config_timestamp,
            "server_timestamp": firestore.SERVER_TIMESTAMP,  # This will be UTC
            "tenant_ref": tenant_reference
        }
    )
    return ref[1]


def _create_jobs(
        db: firestore.Client, jobs: List[Dict], configuration_ref: firestore.DocumentReference
) -> Dict:
    """Creates the jobs in firestore from the manual form collection,
    and return the dict of Job Form ID to Document References"""
    return_dict = {}
    for job in jobs:
        job_dict = job.to_dict()
        job_dict["configuration_ref"] = configuration_ref
        result = db.collection(jobs_collections_name).add(job_dict)
        return_dict[job.id] = result[1]

    return return_dict


def _create_machines(
        db: firestore.Client,
        machines: List[Dict],
        edges_refs: Dict[str, firestore.DocumentReference],
        configuration_ref: firestore.DocumentReference
) -> Dict[str, firestore.DocumentReference]:
    """Creates the Machines entities in the DB based on the received dictionaries.

    It will create the references to the appropriates edges and gateway for each machine.

    Returns:
        Dict of machines IDs to the firestore DocumentReference's for those machines.
    """
    return_dict = {}
    for machine_dict in machines:
        # replace associated_with_edges with edges references
        edges_refs_for_machine = list(
            map(
                lambda edge_logical_id: edges_refs[edge_logical_id],
                machine_dict["associated_with_edges"],
            )
        )
        machine_dict["associated_with_edges"] = edges_refs_for_machine
        machine_dict["configuration_ref"] = configuration_ref

        # creates machine
        log.info(f"Creating Machine in Firestore: {machine_dict['machine_id']}")
        result = db.collection(machines_collection_name).add(machine_dict)
        return_dict[machine_dict["machine_id"]] = result[1]

    return return_dict


def _create_gateways(
        db: firestore.Client,
        gateways: List[Dict],
        created_machines: Dict[str, firestore.DocumentReference],
        created_jobs: Dict[str, firestore.DocumentReference],
        configuration_ref: firestore.DocumentReference,
) -> Dict[str, firestore.DocumentReference]:
    """Creates the new entities for the Gateways based on the received
    dictionaries.

    Returns:
        Returns a Dict of gateway IDs to the firestore DocumentReference for that
        gateway ID.
    """
    return_dict = {}
    for gateway_dict in gateways:
        log.info(f"Creating Gateway in Firestore: {gateway_dict['gateway_id']}")
        gateway_dict["configuration_ref"] = configuration_ref
        gateway_dict["has_job"] = created_jobs[gateway_dict["has_job"]]
        gateway_dict["machines"] = list(
            map(lambda machine: created_machines[machine], gateway_dict["machines"]))  # list(created_machines.values())
        result = db.collection(gateways_collection_name).add(gateway_dict)
        return_dict[gateway_dict["gateway_id"]] = result[1]

    return return_dict


# ======================== Main Function Body to export ============================================
def update_manual_configuration(request):
    """The cloud function body defined in this module, to be deployed
    as a GCP cloud function.

    This method will generate a manual configuration snapshot given the
    tenant-id and the manual configuration collection stored in firestore.

    Args:
        request (flask.Request): The request object.
        <https://flask.palletsprojects.com/en/1.1.x/api/#incoming-request-data>

    Returns:
        200 OK status code if everything was successful.

    """
    print(request)
    request_json = request.get_json(silent=True)
    print(request_json)
    if request_json and 'tenant_id' in request_json:
        tenant_id = request_json['tenant_id']
        configuration_document = request_json['configuration_document']
        log.info(f"Tenant ID received: {tenant_id}")
        log.info(f"Configuration Document in {manual_configuration_collection}"
                 f" collection: {configuration_document}")
    else:
        raise ValueError("JSON is invalid, or missing a 'name' property")

    db = firestore.Client()
    log.info(f"Trying to find Tenant document for {tenant_id}...")

    # Pull the tenant Document
    tenant = db.collection(tenants_collection).where(
        'tenant_id', '==', tenant_id
    ).get()

    if not tenant:
        log.error(f"Tenant document was not found for {tenant_id}")
        return 404

    if len(tenant) > 1:
        log.warning(f"Multiple Tenants in the database with id: {tenant_id}.")

    tenant_ref = tenant[0].reference

    log.info(
        "Connection to Firestore established. Proceed to pulling entities from manual"
        f"configuration form {configuration_document}"
    )

    machines_groups: List[Dict] = pull_docs_from_manual_form_collection(
        db, configuration_document, machines_group_collection_name
    )

    if not machines_groups:
        log.error("No machines_groups documents found. Execution finished")
        return "Error: No machines_groups found\n", 500

    machines_groups

    machines: List[Dict] = pull_docs_from_manual_form_collection(
        db, configuration_document, machines_collection_name)

    if not machines:
        log.error("No machines documents found. Execution finished")
        return "Error: No machines found in database form\n", 500

    gateways: List[Dict] = pull_docs_from_manual_form_collection(
        db, configuration_document,gateways_collection_name)

    if not gateways:
        log.error("No gateways documents found. Execution finished")
        return "Error: No gateways found in database form\n", 500

    devices: List[Dict] = pull_docs_from_manual_form_collection(
        db, configuration_document, devices_collection_name)

    if not devices:
        log.error("No devices documents found. Execution finished")
        return "Error: No devices found in database form\n", 500

    edges: List[Dict] = pull_docs_from_manual_form_collection(
        db, configuration_document, edges_collection_name)

    if not edges:
        log.error("No edges documents found. Execution finished")
        return "Error: No edges found in database form\n", 500

    jobs = pull_docs_from_manual_form_collection(
        db, configuration_document, jobs_collections_name, as_dict=False)

    if not jobs:
        log.error("No jobs document found. Execution finished")
        return "Error: No job entity found in database form\n", 500

    log.info(
        "All the required entities were obtained from the manual configuration form document"
    )

    created_configuration_ref = _create_new_configuration(db, tenant_ref)

    created_devices = txp_firestore_controller._create_devices(db, devices, created_configuration_ref)

    created_edges = txp_firestore_controller._create_edges(db, edges, created_devices, created_configuration_ref)

    created_jobs = _create_jobs(db, jobs, created_configuration_ref)

    created_machines = _create_machines(db, machines, created_edges, created_configuration_ref)

    created_machines_groups = txp_firestore_controller._create_machines_groups(
        db, machines_groups, created_machines, created_configuration_ref
    )

    created_gateways = _create_gateways(db, gateways, created_machines, created_jobs, created_configuration_ref)

    log.info("All the new entities were created for the new configuration snapshot")

    return "OK\n", 200


# ======================== Main program to debug locally ============================================
if __name__ == "__main__":
    from unittest.mock import Mock

    tenant_id = 'labshowroom-001'
    configuration_document = 'laboratorio_showroom_manual_script'
    data = {'tenant_id': tenant_id, 'configuration_document': configuration_document}
    req = Mock(get_json=Mock(return_value=data), args=data)
    update_manual_configuration(req)



# import Dongle_Communication
# import iCOMOX_messages
# import messages_utils
import struct
# import helpers
import math
from common import helpers, iCOMOX_messages, iCOMOX_list, common, messages_utils

# import iCOMOX_list

# once there is a session with the manager, we start the following chain:
# getSystemInfo - to get the manager's macAddress as hardware and software versions
# subscribe - to register to get events and data notifications
#
from communications import Dongle_Communication


class class_iCOMOX_over_Dongle_Communication(Dongle_Communication.class_Dongle_Communication):
    def __init__(self, comm_port, callback_iCOMOX_added=None, callback_iCOMOX_removed=None, callback_updated_iCOMOX_SMIP_list=None, callback_process_message=None, callback_data_sent=None, callback_connection_state=None):# callback_serial_com_open=None, callback_serial_com_close=None)
        Dongle_Communication.class_Dongle_Communication.__init__(
            self=self,
            comm_port=comm_port,
            callback_command_resp=self.on_command_response,
            callback_notif_event=self.on_notif_event,               # notification.events
            callback_notif_log=None,                                # notification.logs
            callback_notif_data=self.on_notif_data,                 # notifications.data - here data from mote should be processed
            callback_notif_ipData=None,                             # notification.ipData - here data from mote using ipv6 should be processed
            callback_notif_healthReport=None,                       # notification.health RawDataSensors
            callback_MgrHello=self.on_manager_session_stop,         # when manager disconnects from us and waiting for ClientHello message (is_active_session -> False)
            callback_HelloResponse=self.on_manager_session_start,   # when manager connects us and waiting for further messages from us (is_active_session -> True)
            callback_serial_comm_open=self.on_serial_comm_open,     # serial COM port is now open
            callback_serial_comm_close=self.on_serial_comm_close    # serial COM port is now closed
        )
        # self.current_iCOMOX_list = []
        # self.temp_iCOMOX_list = []
        self.manager_macAddress = bytearray(8)
        self.sendData_callbackId = 0
        self.callback_updated_iCOMOX_SMIP_list = callback_updated_iCOMOX_SMIP_list
        self.callback_iCOMOX_added = callback_iCOMOX_added
        self.callback_iCOMOX_removed = callback_iCOMOX_removed
        self.callback_process_message = callback_process_message
        self.callback_data_sent = callback_data_sent
        self.callback_connection_state = callback_connection_state

    def send_message_to_iCOMOX(self, macAddress, msg, priority=1, srcPort=0xF0B8, dstPort=0xF0B8):
        return self.send_sendData(macAddress=macAddress, priority=priority, srcPort=srcPort, dstPort=dstPort, data=msg)

    def on_command_response(self, comm, packetType, RC, data):
        if packetType == 0x15:          # reset
            helpers.OUT("Reset ACK, RC={}".format(hex(RC)))
            pass
        elif packetType == 0x16:        # subscribe
            helpers.OUT("subscribe ACK, RC={}".format(hex(RC)))
            self.refresh_current_iCOMOX_list()

        elif packetType == 0x17:        # getTime
            pass
        elif packetType == 0x1A:        # setNetworkConfig
            helpers.OUT("getNetworkConfig ACK, RC={}".format(hex(RC)))
            pass
        elif packetType == 0x1F:        # clearStatistics
            helpers.OUT("clearStatistics ACK, RC={}".format(hex(RC)))
            pass
        elif packetType == 0x21:        # exchangeMoteJoinKey
            pass
        elif packetType == 0x22:        # exchangeNetworkId
            pass
        elif packetType == 0x23:        # radiotestTx
            pass
        elif packetType == 0x25:        # radiotestRx
            pass
        elif packetType == 0x26:        # getRadiotestStatistics
            pass
        elif packetType == 0x27:        # setACLEntry
            pass
        elif packetType == 0x28:        # getNextACLEntry
            pass
        elif packetType == 0x29:        # deleteACLEntry
            pass
        elif packetType == 0x2A:        # pingMote
            helpers.OUT("pingMote ACK, RC={}".format(hex(RC)))
            pass
        elif packetType == 0x2B:        # getLog
            pass
        elif packetType == 0x2C:        # sendData
            if RC == Dongle_Communication.RC_OK:
                [self.sendData_callbackId] = struct.unpack(">L", data)
            else:
                self.sendData_callbackId = 0
            helpers.OUT("sendData ACK, RC={}, callback ID={}".format(hex(RC), hex(self.sendData_callbackId)))

        elif packetType == 0x2D:        # startNetwork (deprecated)
            pass
        elif packetType == 0x2E:        # getSystemInfo
            helpers.OUT("getSystemInfo ACK, RC={}".format(hex(RC)))
            self.manager_macAddress = data[:8]
            hwModel, hwRev, swMajor, swMinor, swPatch, swBuild = struct.unpack_from(">BBBBBH", data, 8)
            #common.app.Information.update_dongle_version(hwModel=hwModel, hwRev=hwRev, swMajor=swMajor, swMinor=swMinor, swPatch=swPatch, swBuild=swBuild, dongle_version_available=True)

            self.send_subscribe(filter=0x12, unackFilter=0xFFFFFFFF)    # subscribe to events (0x02) and data (0x10)

        elif packetType == 0x2F:        # getMoteConfig
            if RC == Dongle_Communication.RC_OK:
                macAddress = data[:8]
                moteID, isAP, state, reserved, isRouting = struct.unpack_from(">HBBBB", data, 8)
                if not isAP:
                    iCOMOX = common.iCOMOXs.add(Type=iCOMOX_list.cCLIENT_TYPE_SMIP, macAddress=macAddress, moteID=moteID, state=state, isRouting=isRouting)
                    if iCOMOX is not None:
                        iCOMOX.state = state
                        iCOMOX.isRouting = isRouting
                        # helpers.OUT("getMoteConfig: failed to add iCOMOX with MAC {}".format(helpers.u8s_to_str(macAddress, ":", "")))
                    self.send_getMoteConfig(macAddress=iCOMOX.macAddress, next=True)
                else:
                    helpers.OUT("Not sending the next getMoteConfig request since we got Access Point mote")

            elif RC == Dongle_Communication.RC_END_OF_LIST:
                # # Copy current_iCOMOX_list[] into temp_iCOMOX_list[]
                # for i in range(0, len(self.temp_iCOMOX_list)):
                #     index = self.index_by_macAddress(macAddress=self.temp_iCOMOX_list[i].macAddress)
                #     if index >= 0:
                #         self.temp_iCOMOX_list[i].packetID = self.current_iCOMOX_list[index].packetID
                #         self.temp_iCOMOX_list[i].received_msg = self.current_iCOMOX_list[index].received_msg
                #         self.temp_iCOMOX_list[i].Reports = self.current_iCOMOX_list[index].Reports
                #         self.temp_iCOMOX_list[i].Hello = self.current_iCOMOX_list[index].Hello
                #         self.temp_iCOMOX_list[i].Name = self.current_iCOMOX_list[index].Name
                #         self.temp_iCOMOX_list[i].UniqueID = self.current_iCOMOX_list[index].UniqueID
                #         self.temp_iCOMOX_list[i].board_type = self.current_iCOMOX_list[index].board_type
                #         # in this copy, state, moteID & isRouting might change due to new information in temp_iCOMOX_list[i]
                #
                # # Exchange between the current_iCOMOX_list & temp_iCOMOX_list without copying anything
                # self.current_iCOMOX_list, self.temp_iCOMOX_list = self.temp_iCOMOX_list, self.current_iCOMOX_list
                # self.temp_iCOMOX_list.clear()

                if callable(self.callback_updated_iCOMOX_SMIP_list):
                    self.callback_updated_iCOMOX_SMIP_list()
                else:
                    helpers.OUT("NO call to callback_updated_iCOMOX_SMIP_list()")
                helpers.OUT("getMoteConfig: RC=RC_END_OF_LIST")
            else:
                helpers.OUT("getMoteConfig RC={}".format(hex(RC)))

        elif packetType == 0x30:        # getPathInfo
            pass
        elif packetType == 0x31:        # getNextPathInfo
            pass
        elif packetType == 0x32:        # setAdvertising
            pass
        elif packetType == 0x33:        # setDownstreamFrameMode
            pass
        elif packetType == 0x35:        # getManagerStatistics
            pass
        elif packetType == 0x36:        # setTime
            pass
        elif packetType == 0x37:        # getLicense (deprecated)
            pass
        elif packetType == 0x38:        # setLicense (deprecated)
            pass
        elif packetType == 0x3A:        # setCLIUser
            pass
        elif packetType == 0x3B:        # sendIP
            pass
        elif packetType == 0x3D:        # restoreFactoryDefaults
            pass
        elif packetType == 0x3E:        # getMoteInfo
            pass
        elif packetType == 0x3F:        # getNetworkConfig
            pass
        elif packetType == 0x40:        # getNetworkInfo
            helpers.OUT("getNetworkInfo ACK, RC={}".format(hex(RC)))
            pass
        elif packetType == 0x41:        # getMoteConfigById
            helpers.OUT("getMoteConfiguById ACK, RC={}".format(hex(RC)))
            pass
        elif packetType == 0x42:        # setCommonJoinKey
            pass
        elif packetType == 0x43:        # getIPConfig
            pass
        elif packetType == 0x44:        # setIPConfig
            pass
        elif packetType == 0x45:        # deleteMote
            helpers.OUT("deleteMote ACK, RC={}".format(hex(RC)))
            pass
        elif packetType == 0x46:        # getMoteLinks
            pass

    # def delete_mote_from_list(self, macAddress):
    #     index = self.index_by_macAddress(macAddress=macAddress)
    #     if index >= 0:
    #         del self.current_iCOMOX_list[index]
    #         if callable(self.callback_iCOMOX_removed):
    #             self.callback_iCOMOX_removed(macAddress=macAddress)

    # def delete_all_motes_from_list(self):
    #     if callable(self.callback_iCOMOX_removed):
    #         for i in range(len(self.current_iCOMOX_list)-1, -1, -1):
    #             macAddress = self.current_iCOMOX_list[i].macAddress
    #             self.callback_iCOMOX_removed(macAddress=macAddress)
    #     del self.current_iCOMOX_list[:]

    # def add_mote_to_list(self, macAddress):
    #     index = self.index_by_macAddress(macAddress=macAddress)
    #     if index < 0:
    #         self.send_getMoteConfig(macAddress=macAddress, next=False)
    #     else:
    #         if callable(self.callback_iCOMOX_added):
    #             self.callback_iCOMOX_added(self.current_iCOMOX_list[index])

    # def add_all_motes_to_list(self):
    #     if callable(self.callback_iCOMOX_added):
    #         for iCmxSmip in self.current_iCOMOX_list:
    #             self.callback_iCOMOX_added(iCmxSmip)

    def change_iCOMOX_SMIP_state(self, macAddress, state=None):
        try:
            iComox = common.iCOMOXs.find_by_macAddress(macAddress=macAddress)
            if iComox is not None:
                if state is not None:
                    iComox.state = state
                if callable(self.callback_updated_iCOMOX_SMIP_list):
                    self.callback_updated_iCOMOX_SMIP_list()
            else:
                helpers.OUT("Failed to find MAC: {}".format(helpers.u8s_to_str(macAddress, ":", "")))
                self.send_getMoteConfig(macAddress=macAddress, next=False)
        except Exception as ex:
            if hasattr(ex, 'message'):
                helpers.OUT(ex.message)
        finally:
            pass

    def on_notif_event(self, comm, eventId, eventType, eventData):
        if eventType == 0:              # moteReset
            macAddress=eventData[0:8]
            helpers.OUT("DeleteMote MAC={}".format(helpers.u8s_to_str(macAddress, ":", "")))
            self.change_iCOMOX_SMIP_state(macAddress=macAddress, state=Dongle_Communication.MOTE_STATE_LOST)
            # self.refresh_current_iCOMOX_list()

        elif eventType == 1:            # networkReset
            common.iCOMOXs.delete_all_of_type(Type=iCOMOX_list.cCLIENT_TYPE_SMIP)
            common.app.Clients.delete_all()
            # self.refresh_current_iCOMOX_list()

        elif eventType == 2:            # commandFinish
            pass

        elif eventType == 3:            # moteJoin
            macAddress = eventData[0:8]
            helpers.OUT("moteJoin MAC: {}".format(helpers.u8s_to_str(macAddress, ":", "")))
            self.change_iCOMOX_SMIP_state(macAddress=macAddress, state=Dongle_Communication.MOTE_STATE_NEGOTIATING)

        elif eventType == 4:            # moteOperational
            macAddress = eventData[0:8]
            helpers.OUT("moteOperational MAC: {}".format(helpers.u8s_to_str(macAddress, ":", "")))
            self.change_iCOMOX_SMIP_state(macAddress=macAddress, state=Dongle_Communication.MOTE_STATE_OPERATIONAL)

        elif eventType == 5:            # moteLost
            macAddress = eventData[0:8]
            helpers.OUT("moteLost MAC: {}".format(helpers.u8s_to_str(macAddress, ":", "")))
            self.change_iCOMOX_SMIP_state(macAddress=macAddress, state=Dongle_Communication.MOTE_STATE_LOST)

        elif eventType == 6:            # networkTime
            pass
        elif eventType == 7:            # pingResponse
            pass
        elif eventType == 10:           # pathCreate
            pass
        elif eventType == 11:           # pathDelete
            pass
        elif eventType == 12:           # packetSent
            # Triggered by successful sendData command
            if callable(self.callback_data_sent):
                sendData_callbackId, RC = struct.unpack(">LB", eventData)
                self.callback_data_sent(self.Comm_LTC5800IPR, RC, sendData_callbackId == self.sendData_callbackId)
            self.sendData_callbackId = 0

        elif eventType == 13:           # moteCreate
            macAddress = eventData[0:8]
            [moteID] = struct.unpack_from(">H", eventData, 8)
            helpers.OUT("moteCreate MAC: {}".format(helpers.u8s_to_str(macAddress, ":", "")))
            iComox = common.iCOMOXs.add(Type=iCOMOX_list.cCLIENT_TYPE_SMIP, macAddress=macAddress, isRouting=False, moteID=moteID, state=Dongle_Communication.MOTE_STATE_LOST)
            # self.change_iCOMOX_SMIP_state(macAddress=macAddress, state=None)

        elif eventType == 14:           # moteDelete
            macAddress = eventData[0:8]
            helpers.OUT("moteDelete MAC: {}".format(helpers.u8s_to_str(macAddress, ":", "")))
            self.change_iCOMOX_SMIP_state(macAddress=macAddress, state=Dongle_Communication.MOTE_STATE_LOST)

        elif eventType == 15:           # joinFailed
            macAddress = eventData[0:8]
            helpers.OUT("joinFailed MAC: {}, due to reason {}".format(helpers.u8s_to_str(macAddress, ":", ""), eventData[8]))
            self.change_iCOMOX_SMIP_state(macAddress=macAddress, state=Dongle_Communication.MOTE_STATE_NEGOTIATING)

        elif eventType == 16:           # invalidMIC
            macAddress = eventData[0:8]
            helpers.OUT("InvalidMIC MAC: {}".format(helpers.u8s_to_str(macAddress, ":", "")))
            self.change_iCOMOX_SMIP_state(macAddress=macAddress, state=Dongle_Communication.MOTE_STATE_LOST)

    def on_notif_data(self, comm, timestamp, macAddress, srcPort, dstPort, data):
        iComox = common.iCOMOXs.find_by_macAddress(macAddress=macAddress)
        if (iComox is None) or (len(data) < iCOMOX_messages.SMIP_PACKET_HEADER_SIZE + 1):  # = sizeof(packet index) + at least 1 byte for data
            return

        packet_index = data[0]
        packet_data = data[iCOMOX_messages.SMIP_PACKET_HEADER_SIZE:]

        if packet_index == 0:    # start a new message
            if (packet_data[0] < iCOMOX_messages.cCOMOX_MSG_CODE_COUNT) and (packet_data[0] != iCOMOX_messages.cCOMOX_MSG_CODE_Report):
                helpers.OUT("Get message ({}) from {}".format(hex(packet_data[0]), helpers.u8s_to_str(macAddress, ":", "")))
            iComox.accumulated_msg = bytearray()
            iComox.expected_msg_size = 0
            iComox.packetID = 0
            # if (packet_data[0] == iCOMOX_messages.cCOMOX_MSG_CODE_Report) and (len(packet_data) < iCOMOX_messages.COMOX_IN_MSG_SIZE[iCOMOX_messages.cCOMOX_MSG_CODE_Report]):
            #     helpers.OUT("on_notify_data: packet 0 contains less bytes than report message header")
            #     return
            # Determine the incoming message expected size, assuming packet 0 is enough for determining it
            while len(packet_data) > 0:
                bytes_to_read = messages_utils.on_get_in_message_size(accumulated_msg=iComox.accumulated_msg, adxl356_smip=True)
                if bytes_to_read < 0:   # check if the message is a legal one
                    helpers.OUT("on_notif_data: Illegal message was received")
                    if iComox == common.iCOMOXs.current:
                        common.app.StatusBar.set("Unrecognized message received")
                        common.app.StatusBar.progressbar_hide()
                    return
                elif bytes_to_read > 0:
                    iComox.expected_msg_size = len(iComox.accumulated_msg) + bytes_to_read
                    bytes_read = min(len(packet_data), bytes_to_read)
                    iComox.accumulated_msg += packet_data[:bytes_read]
                    packet_data = packet_data[bytes_read:]
                    # bytes_to_read -= bytes_read
                else:
                    break

        else:
            if packet_index != iComox.packetID:   # if we got an unexpected packetID in the packet
                if iComox == common.iCOMOXs.current:
                    common.app.StatusBar.set("Unexpected packet ID (received: {}, expected: {})".format(packet_index+1, iComox.packetID+1))
                    common.app.StatusBar.progressbar_hide()
                iComox.packetID = 0  # mark that the next packet must start a new message (packetID = 0)
                return

            # Add the received packet to the accumulated message
            iComox.accumulated_msg += packet_data[0:min(len(packet_data), Dongle_Communication.SMIP_MAX_PAYLOAD_TO_MANAGER-iCOMOX_messages.SMIP_PACKET_HEADER_SIZE, iComox.expected_msg_size - len(iComox.accumulated_msg))]

        # Display the message progress in statusBar
        max_num_of_packets = int(math.ceil(iComox.expected_msg_size/(Dongle_Communication.SMIP_MAX_PAYLOAD_TO_MANAGER-iCOMOX_messages.SMIP_PACKET_HEADER_SIZE)))
        if iComox == common.iCOMOXs.current:
            msg_description = messages_utils.get_IN_MSG_Description(msg=iComox.accumulated_msg)
            s = 'Received packet {} out of {} ({}) from MAC address {}'.format(packet_index+1, max_num_of_packets, msg_description, helpers.u8s_to_str(macAddress, ":"))
            common.app.StatusBar.set(s)
            common.app.StatusBar.message_progress(packet_index=packet_index+1, total_packets=max_num_of_packets)
            #helpers.OUT(s)

        if len(iComox.accumulated_msg) >= iComox.expected_msg_size:
            iComox.accumulated_msg = iComox.accumulated_msg[:iComox.expected_msg_size]
            if callable(self.callback_process_message):
                self.callback_process_message(msg=iComox.accumulated_msg, iComox=iComox)
            iComox.packetID = 0    # mark that the next packet must start a new message (packetID = 0)
            if iComox == common.iCOMOXs.current:
                common.app.StatusBar.message_progress(packet_index=0, total_packets=max_num_of_packets)
        else:
            iComox.packetID += 1

    def on_manager_session_stop(self, comm, mgrHelloCounter):
        if (mgrHelloCounter >= 3):
            comm.send_ClientHello()
        if self.callback_connection_state is not None:
            self.callback_connection_state(serial_connection=True, session=False)

    def on_manager_session_start(self, comm, RC, version, mode):
        self.send_getSystemInfo()
        if self.callback_connection_state is not None:
            self.callback_connection_state(serial_connection=True, session=True)

    def on_serial_comm_open(self, comm):
        if self.callback_connection_state is not None:
            self.callback_connection_state(serial_connection=True, session=False)

    def on_serial_comm_close(self, comm):
        if self.callback_connection_state is not None:
            self.callback_connection_state(serial_connection=False, session=False)

    def refresh_current_iCOMOX_list(self):
        # self.temp_iCOMOX_list.clear()
        return self.send_getMoteConfig(macAddress=self.manager_macAddress, next=True)
    #
    # def index_by_macAddress(self, macAddress):
    #     # return [i if macAddress == x.macAddress else -1 for i, x in enumerate(self.current_iCOMOX_list)]
    #     for i in range(0, len(self.current_iCOMOX_list)):
    #         if self.current_iCOMOX_list[i].macAddress == macAddress:
    #             return i
    #     return -1
    #
    # def index_by_moteId(self, moteId):
    #     # return [i if moteId == x.moteId else -1 for i, x in enumerate(self.current_iCOMOX_list)]
    #     for i in range(0, len(self.current_iCOMOX_list)):
    #         if self.current_iCOMOX_list[i].moteId == moteId:
    #             return i
    #     return -1




import logging

import numpy as np
import scipy.signal as signal
from scipy.integrate import cumtrapz
import sys


def peak(data):
    return np.absolute(data).max()


def rms(data):
    return np.sqrt(np.power(data, 2).mean())


def standard_deviation(data):
    return np.std(data)


def crest_factor(data):
    cf = peak(data) / rms(data)
    if np.isnan(cf):
        cf = sys.float_info.max
        logging.warning("crest factor to high returning maximum possible value")
    return cf


def get_psd(data, sampling_frequency, window="hanning"):
    bin_width = sampling_frequency / len(data)
    return signal.welch(data, fs=sampling_frequency, nperseg=sampling_frequency / bin_width, window=window, axis=0)


def integrate(y, x=None, sampling_frequency=1):
    if x is not None:
        return cumtrapz(y, x=x)
    else:
        return cumtrapz(y, dx=sampling_frequency)






# =========================== Imports ===============================
import os
import json
import base64
from google.cloud import storage
from txp.common.protos.gateway_package_pb2 import GatewayPackageProto
import google.cloud.firestore as firestore
from txp.cloud.common import utils as cu

# =========================== Constant Data & helper functions ===============================
project_id = os.environ.get('GCP_PROJECT')
data_lake_bucket_name = \
    f"{project_id}-{os.environ.get('DATA_LAKE_BUCKET_NAME', 'tranxpert-mvp-telemetry-data-lake')}"
pending_signal_chunks_collection_name = os.environ.get('PENDING_SIGNAL_CHUNKS_COLLECTION',
                                                       'pending_signal_chunks')


# =========================== Cloud function def ===============================

def insert_pending_signal_chunk(chunk_id, gateway_id, db):
    db.collection(pending_signal_chunks_collection_name).document(chunk_id).set({
        "id": chunk_id,
        "path": f"{gateway_id}/{chunk_id}"
    })


def get_chunk_id(chunk):
    return (
        f"""{chunk["edge_logical_id"]}|"""
        f"""{chunk["perception_name"]}|"""
        f"""{chunk["package_timestamp"]}|"""
        f"""{chunk["signal_timestamp"]}|"""
        f"""{chunk["previous_part_index"]}|"""
        f"""{chunk["part_index"]}"""
    )


def store_telemetry_event_to_data_lake(event, context):
    """This function is the function deployed to run in the cloud
    functions runtime.
    A detailed page with all the context can be found here:
        https://cloud.google.com/functions/docs/calling/pubsub
    In order to know what exactly this function does take a look at readme.md file
    """
    print('Starting to process telemetry event received...')
    if 'data' in event:
        if 'attributes' not in event:
            print('Wrong telemetry event received stopping function execution')
            return
        telemetry_data = base64.b64decode(event['data']).decode()
        proto_string = base64.b64decode(telemetry_data)
        proto = GatewayPackageProto()
        proto.ParseFromString(proto_string)
        chunks = cu.from_proto_to_json(proto)
        gateway_id = event['attributes']['gatewayId']
        storage_client = storage.Client()
        bucket = storage_client.bucket(data_lake_bucket_name, user_project=project_id)
        firestore_db = firestore.Client()
        for chunk in chunks:
            print(f"Processing chunk: {get_chunk_id(chunk)}")
            blob = bucket.blob(f"{gateway_id}/{get_chunk_id(chunk)}")
            blob.upload_from_string(
                data=json.dumps(chunk),
                content_type='application/json'
            )
            insert_pending_signal_chunk(get_chunk_id(chunk), gateway_id, firestore_db)
        firestore_db.close()



import click
from txp.devices.drivers.cli_commands.test_voyager import voyager_test
from txp.common.config import settings
from txp.main import main
import txp.web.run_app import run_st_app
import multiprocessing
from sys import platform
import logging
log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


@click.group()
def txp_cli_entrypoint():
    """TXP command line for Tranxpert predictive maintenance solution."""
    pass


@click.command()
@click.argument("device-mac", type=str)
def run_voyager_test(device_mac):
    """Run voyager test

    DEVICE_MAC is the Voyager device string MAC address in Dash-Hexadecimal notation.
    """
    mac_tpl = tuple(device_mac)
    voyager_test(mac_tpl)


@click.command()
def web():
    """Run the streamlit web application"""
    run_st_app()


@click.command()
@click.option('--clear-spilled-packages', is_flag=True, default=False,
              help="If enabled, then the packages found on disk will be discarded.")
@click.option('--packages-queue-memory-limit', default=settings.gateway.queue_memory_size,
              help="Percentage of memory size at which the "
                   "gateway will start to spill packages. Is set to 0, then the Gateway"
                   " will preserver all the sampled packages and it'll operate without "
                   " sending the packages to the cloud.")
# TODO: Only should be necessary 1 topic value and the streaming mode.
@click.option('--telemetry-topic-realtime-template',
              default=settings.gateway.mqtt.telemetry_topic_realtime_template,
              help='The telemetry topic template string to use when sending telemetry '
                   'messages in realtime')
@click.option('--telemetry-topic-batch-template',
              default=settings.gateway.mqtt.telemetry_topic_batch_template,
              help="The telemetry topic template string to use when sending telemetry"
                   "messages in realtime")
@click.option('--streaming-mode', default=settings.gateway.streaming_mode,
              help="The mode to use for telemetry: REALTIME / BATCH")
def run(
        clear_spilled_packages: bool,
        packages_queue_memory_limit: int,
        telemetry_topic_realtime_template: str,
        telemetry_topic_batch_template: str,
        streaming_mode: str
):
    """Run the main application for the Gateway"""

    # Sets the packages queue size to default/received value
    settings.gateway.queue_memory_size = int(packages_queue_memory_limit)

    if packages_queue_memory_limit == 0:
        log.info("--packages-queue-memory-limit equal to 0.  "
                  "The Gateway will operate in offline mode after initialization")
        settings.gateway.queue_memory_size = 0
        settings.gateway.offline_mode = True

    if clear_spilled_packages:
        log.info("--clear-spilled-packages flag enabled. "
                  "Old packages will be deleted.")
        settings.gateway.queue_clean_old_spilled_packages = True

    if telemetry_topic_realtime_template:
        log.info(f"Set telemetry realtime topic template to: {telemetry_topic_realtime_template}")
        settings.gateway.mqtt.telemetry_topic_realtime_template = telemetry_topic_realtime_template

    if telemetry_topic_batch_template:
        log.info(f"Set telemetry batch topic template to: {telemetry_topic_batch_template}")
        settings.gateway.mqtt.telemetry_topic_batch_template = telemetry_topic_batch_template

    if streaming_mode:
        log.info(f"Streaming mode setup to: {streaming_mode}")
        settings.gateway.streaming_mode = streaming_mode

    main()


txp_cli_entrypoint.add_command(run_voyager_test)
txp_cli_entrypoint.add_command(run)
txp_cli_entrypoint.add_command(web)


if __name__ == '__main__':

    # This avoids opencv conflicts that can arise when using multiprocessing with
    # fork in macOS.
    if platform == "Darwin" or platform == "darwin":
        multiprocessing.set_start_method('spawn', force=True)
    run()



import time

from google.cloud import pubsub_v1
from google.api_core import retry
from google.oauth2 import service_account
import google.cloud.firestore as firestore
from txp.cloud.common import utils as cu
from txp.cloud.controllers import bigquery as bq
import json
from google.cloud import bigquery
from txp.cloud import settings
import logging


class PubSubSubscriber:
    """Subscriber at GCP PubSub, it reads messages from PubSub service, this class reads messages by  batches of
    max_messages length.

    Args:
        credentials: GCS credentials.
        subscription_id: PubSub subscription id where signal messages are going to come from.
    """

    max_messages = 20
    max_retries = 10

    def __init__(self, credentials, subscription_id):
        self.project_id = credentials.project_id
        self.subscription_id = subscription_id
        self.subscriber = pubsub_v1.SubscriberClient(credentials=credentials)

    def get_messages(self):
        subscription_path = self.subscriber.subscription_path(self.project_id, self.subscription_id)
        with self.subscriber:
            response = self.subscriber.pull(
                request={"subscription": subscription_path, "max_messages": self.max_messages},
                retry=retry.Retry(deadline=300),
            )

            if len(response.received_messages) == 0:
                return

            ack_ids = []
            for received_message in response.received_messages:
                ack_ids.append(received_message.ack_id)

            self.subscriber.acknowledge(request={"subscription": subscription_path, "ack_ids": ack_ids})

        return response.received_messages


class SignalsSegregator:
    """This class is in charge of waiting for all necessary PubSub messages for doing a prediction, due to
    telemetry pipelines PubSub messages do not come in order, prediction service has to wait for all messages
    generated at the pipeline in order to make a prediction for the signal.

    Args:
        credentials_str: GCS credentials string.
        dataset: BQ dataset name.
        subscription_id: PubSub subscription id where signal messages are going to come from.
        backup_collection_name: Firestore collection name where all backup is going to be stored.
    """

    def __init__(self, credentials_str, dataset, subscription_id, backup_collection_name, log_level=logging.INFO):
        logging.basicConfig(level=log_level)
        json_dict_service_account = json.loads(credentials_str, strict=False)
        self.credentials = service_account.Credentials.from_service_account_info(json_dict_service_account)
        self.subscription_id = subscription_id
        self.firestore_db = firestore.Client(credentials=self.credentials, project=self.credentials.project_id)
        self.backup_collection_name = backup_collection_name
        self.messages_db = self.load_backup()
        self.dataset = dataset
        self.bigquery_segregator = BigquerySignalsSegregator(credentials_str, dataset)

    def load_backup(self):
        logging.info("Loading signals segregator backup")
        backup_ref = self.firestore_db.collection(self.backup_collection_name).document(
            settings.firestore.ray_signals_segregator_doc_name
        )
        backup = backup_ref.get().to_dict()
        backup_ref.set({})

        if backup:
            logging.info("There were signals at firestore db")
        else:
            logging.info("There were no signals at firestore db")

        logging.info("Loaded signals segregator")

        return backup

    def backup(self):
        logging.info("Backing up signals segregator")
        self.firestore_db.collection(self.backup_collection_name).document(
            settings.firestore.ray_signals_segregator_doc_name
        ).set(self.messages_db)
        logging.info("Signals segregator backed up")

    def __store_message(self, message, schema, task_id):
        table_id = message["table_id"].split(".")[2]
        observation_timestamp = str(message["observation_timestamp"])
        try:
            x = schema[message["edge_logical_id"]][message["perception_name"]][table_id]
        except KeyError:
            return False

        task = self.messages_db.get(task_id, None)
        if task is None:
            self.messages_db[task_id] = {}

        window = self.messages_db[task_id].get(observation_timestamp, None)
        if window is None:
            empty_doc = {}
            for edge_logical_id in schema:
                empty_doc[edge_logical_id] = {}
                for perception in schema[edge_logical_id]:
                    empty_doc[edge_logical_id][perception] = {}
                    for table in schema[edge_logical_id][perception]:
                        if schema[edge_logical_id][perception][table]:
                            empty_doc[edge_logical_id][perception][table] = []
            window = empty_doc
            self.messages_db[task_id][observation_timestamp] = window

        self.messages_db[task_id][observation_timestamp][message["edge_logical_id"]][
            message["perception_name"]][table_id].append(message)
        return True

    def __check_metrics_table_completeness(self, rows):
        if not len(rows):
            return False
        configuration_id = rows[0]["configuration_id"]
        edge_logical_id = rows[0]["edge_logical_id"]
        perception_name = rows[0]["perception_name"]
        tenant_id = rows[0]["tenant_id"]
        return len(rows) == cu.get_signal_dimensions_from_firestore(configuration_id, tenant_id, edge_logical_id,
                                                                    perception_name, self.firestore_db)[0]

    @staticmethod
    def __check_time_table_completeness(rows):
        rows = sorted(rows, key=lambda d: d['part_index'])
        return len(rows) and rows[0]["previous_part_index"] >= rows[0]["part_index"] and len(rows) == (
                rows[0]["previous_part_index"] - rows[0]["part_index"] + 1)

    def __check_sampling_window_completeness(self, sampling_window):
        complete = True
        for edge_logical_id in sampling_window:
            for perception_name in sampling_window[edge_logical_id]:
                for table in sampling_window[edge_logical_id][perception_name]:
                    elements = sampling_window[edge_logical_id][perception_name][table]
                    if table == "time":
                        complete = complete and self.__check_time_table_completeness(elements)
                    elif table == "time_metrics" or table == "fft_metrics" or table == "psd_metrics":
                        complete = complete and self.__check_metrics_table_completeness(elements)
                    else:
                        complete = complete and len(elements)
        return complete

    def get_messages(self):

        messages = None
        for i in range(0, PubSubSubscriber.max_retries):
            try:
                subscriber = PubSubSubscriber(self.credentials, self.subscription_id)
                messages = subscriber.get_messages()
                break
            except Exception as e:
                logging.info(f"Unexpected error when reading pubsub Messages, retrying Exception: {e}")
                if i + 1 == PubSubSubscriber.max_retries:
                    raise e
                time.sleep(10)
        return messages

    def process_message(self, message, schema, task_id):
        is_valid_message = self.__store_message(message, schema, task_id)
        if not is_valid_message:
            return False

        sampling_window_data = self.messages_db[task_id][str(message["observation_timestamp"])]
        completed = self.__check_sampling_window_completeness(sampling_window_data)

        return completed

    def get_sampling_window(self, observation_timestamp, task_id):
        return self.messages_db[task_id][str(observation_timestamp)]

    def delete_sampling_window(self, observation_timestamp, task_id):
        self.messages_db[task_id].pop(str(observation_timestamp))
        if not self.messages_db[task_id]:
            self.messages_db.pop(task_id)


class BigquerySignalsSegregator:

    def __init__(self, credentials_str, dataset):
        self.credentials_str = credentials_str
        self.dataset = dataset

    @staticmethod
    def __build_time_signal(rows):
        signal = cu.merge_signal_chunks(rows)
        signal["data"] = [dimension["values"] for dimension in signal["data"]]
        return signal

    def get_sampling_window_rows_from_bigquery(self, sampling_window, observation_timestamp, tenant_id):
        json_dict_service_account = json.loads(self.credentials_str, strict=False)
        credentials = service_account.Credentials.from_service_account_info(json_dict_service_account)
        bigquery_client = bigquery.Client(credentials=credentials, project=credentials.project_id)
        sampling_window_rows = {}
        for edge_logical_id in sampling_window:
            sampling_window_rows[edge_logical_id] = {}
            for perception_name in sampling_window[edge_logical_id]:
                sampling_window_rows[edge_logical_id][perception_name] = {}
                for table_id in sampling_window[edge_logical_id][perception_name]:
                    sampling_window_rows[edge_logical_id][perception_name][
                        table_id] = bq.get_sampling_window_from_bigquery(
                        f'{self.dataset}.{table_id}', edge_logical_id, perception_name, observation_timestamp,
                        tenant_id, bigquery_client)
                    if table_id == "time":
                        sampling_window_rows[edge_logical_id][perception_name][table_id] = [
                            self.__build_time_signal(sampling_window_rows[edge_logical_id][perception_name][table_id])]
        bigquery_client.close()
        return sampling_window_rows
